/*                               -*- Mode: C -*- 
 * linux-mdep.c --- Linux kernel dependent support for LiS.
 * Author          : Francisco J. Ballesteros
 * Created On      : Sat Jun  4 20:56:03 1994
 * Last Modified By: John A. Boyd Jr.
 * RCS Id          : $Id: linux-mdep.c,v 1.8 1996/01/26 16:53:01 dave Exp $
 * Purpose         : provide Linux kernel <-> LiS entry points.
 * ----------------______________________________________________
 *
 *    Copyright (C) 1995  Francisco J. Ballesteros, Denis Froschauer
 *    Copyright (C) 1997-2000
 *			  David Grothe, Gcom, Inc <dave@gcom.com>
 *    Copyright (C) 2000  John A. Boyd Jr.  protologos, LLC
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 *
 *    You can reach me by email to
 *    nemo@ordago.uc3m.es, 100741.1151@compuserve.com
 *    dave@gcom.com
 */

#ident "@(#) LiS linux-mdep.c 2.42 5/7/01 21:36:35 "

/*  -------------------------------------------------------------------  */
/*				 Dependencies                            */

#include <sys/LiS/linux-mdep.h>
#include <sys/lislocks.h>
#include <linux/module.h>
#include <linux/version.h>

#ifdef USE_BH
#include <linux/stddef.h>		/* for NULL */
#include <linux/tqueue.h>		/* task queues */
#include <asm/ptrace.h>			/* for pt_regs */
#include <linux/interrupt.h>
#else
#include <linux/sched.h>
#define	__KERNEL_SYSCALLS__	1	/* to make kernel_thread visible */
#include <linux/unistd.h>		/* for kernel_thread */
#endif

#include <asm/signal.h>
#include <asm/io.h>
#include <sys/strport.h>	/* interface */
#include <sys/LiS/errmsg.h>	/* LiS err msg types */
#include <sys/LiS/buffcall.h>	/* bufcalls */
#include <sys/LiS/head.h>	/* stream head */
#include <sys/osif.h>
#include <sys/stream.h>		/* LiS entry points */
#include <sys/cmn_err.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/sysmacros.h>
#include <linux/mm.h>
#include <linux/file.h>
#include <linux/fs.h>		/* linux file sys externs */
#ifdef KERNEL_2_3
#include <linux/mount.h>
#endif
#include <linux/time.h>

#if 0 /* OSH: Now using lis_module_info, and a #define in <sys/LiS/mod.h> */
#ifdef KERNEL_2_1		/* kernel version 2.1 or later */
#define module_info	xx_module_info
#endif
#endif

/*  -------------------------------------------------------------------  */
/*  -------------------------------------------------------------------  */

int lis_major;
char lis_kernel_version[]= UTS_RELEASE;


extern char	lis_version[] ;
extern char	lis_date[] ;

extern void     hook_syscalls(void);
extern void     restore_syscalls(void);

char	*lis_stropts_file =
#if   defined(_LIS_STROPTS_H)
                "<LiS/include/stropts.h>"
#elif defined(_LIS_SYS_STROPTS_H)
		"<LiS/include/sys/stropts.h>"
#elif defined(_STROPTS_H)
                "<usr/include/stropts.h>"
#elif defined(_SYS_STROPTS_H)
                "<usr/include/sys/stropts.h>"
#elif defined(_BITS_STROPTS_H)
		"<usr/include/bits/stropts.h>"
#else
		"<unknown/stropts.h>"
#endif
;

/************************************************************************
*                            lis_assert_fail                            *
*************************************************************************
*									*
* This is called when the ASSERT() macro fails.                         *
*									*
* It sends a warning message to the kernel logger, but does nothing     *
* else.                                                        		*
*									*
************************************************************************/
void lis_assert_fail(const char *expr, const char *objname,
		     const char *file, unsigned int line)
{
        printk(KERN_CRIT "%s: assert(%s) failed in file %s, line %u\n",
	       objname, expr, file, line);
        /* We cannot just abort() the kernel here :-( */
}

/************************************************************************
*                            Prototypes                                 *
************************************************************************/
extern void do_gettimeofday(struct timeval *tv) ;	/* kernel fcn */
extern void lis_print_trace_func(char *msg) ;	/* forward decl */
extern void lis_spl_init(void);			/* lislocks.c */

/************************************************************************
*                          vremap                                       *
*************************************************************************
*									*
* The 2.1 and 2.2 kernels have de-implemented the function vremap.	*
* They replaced it with ioremap, which does the same thing.  For	*
* backward compatibility we provide a vremap that calls ioremap.	*
*									*
************************************************************************/
#if 0				/* now done via osif.[ch] */
#ifdef KERNEL_2_1		/* kernel version 2.1 or later */

void *vremap(unsigned long offset, unsigned long size)
{
    return(ioremap(offset, size)) ;
}

#endif
#endif

/************************************************************************
*                           lis_phys_to_virt                            *
*************************************************************************
*									*
* This routine maps physical addresses to virtual addresses and is	*
* portable across kernel versions.					*
*									*
* This may not look like much, but the secret is that the routine	*
* phys_to_virt is an inline function that is different in the different	*
* kernel versions.  Because we do this here, driver writers do not	*
* necessarily have to maintain a special version of their drivers	*
* for kernel version 2.2.						*
*									*
************************************************************************/
#if 0				/* now done via osif.[ch] */
void *lis_phys_to_virt(unsigned long address)
{
    return(phys_to_virt(address)) ;
}
#endif

/************************************************************************
*                       Storage Declarations                            *
************************************************************************/

lis_spin_lock_t			lis_runq_lock ;	/* syncs global operations */
lis_spin_lock_t			lis_setqsched_lock ; /* one qsched at a time */
lis_semaphore_t			lis_runq_sems[LIS_NR_CPUS] ;
lis_semaphore_t			lis_runq_kill_sems[LIS_NR_CPUS] ;
int				lis_runq_sched ;     /* q's are scheduled */
lis_atomic_t			lis_inode_cnt ;
lis_spin_lock_t			lis_task_lock ; /* for creds operations */

/*
 * Declared in head.c
 */
extern lis_atomic_t		lis_runq_cnt ;	/* # of threads running */
extern int			lis_num_cpus ;
extern int			lis_runq_pids[LIS_NR_CPUS] ;
extern lis_atomic_t		lis_runq_active_flags[LIS_NR_CPUS] ;
extern volatile unsigned long	lis_setqsched_cnts[LIS_NR_CPUS] ;
extern volatile unsigned long	lis_setqsched_isr_cnts[LIS_NR_CPUS] ;
extern void			lis_run_queues(int) ;
extern lis_atomic_t		lis_queues_running ;
extern lis_atomic_t		lis_putnext_flag ;
extern lis_atomic_t	 	lis_runq_req_cnt ;
extern lis_spin_lock_t	 	lis_qhead_lock ;


/*  -------------------------------------------------------------------  */

/* This should be entry points from the kernel into LiS
 * kernel should be fixed to call them when appropriate.
 */

/*
 * File operations
 */
struct file_operations
lis_streams_fops = {
    read:      lis_strread,		/* read    		*/
    write:     lis_strwrite,		/* write                */
#if defined(KERNEL_2_1)
    poll:      lis_poll_2_1,		/* poll  		*/
#else
    select:    lis_select,		/* select  		*/
#endif
    ioctl:     lis_strioctl,		/* ioctl   		*/
    open:      lis_stropen,		/* open                 */
    release:   lis_strclose,		/* release 		*/
};

/*
 * The authors and maintainers of LiS would like to thank Al Viro
 * for making necessary the following 60 or so lines of code,
 * together with the supporting routines farther down in this file.
 * It is one of the more complex implementations of a NO-OP that
 * you are likely to see outside of the Windows NT kernel.
 */
/*
 * Dentry operations
 */
#if defined(KERNEL_2_3)
extern int lis_delete_dentry(struct dentry *dentry) ;

struct dentry_operations lis_dentry_ops =
{
    d_delete:	lis_delete_dentry
};
#elif defined(KERNEL_2_1)
extern void lis_delete_dentry(struct dentry *dentry) ;

struct dentry_operations lis_dentry_ops =
{
    d_delete:	lis_delete_dentry
};
#endif

/*
 * Inode operations
 */
struct inode_operations
lis_streams_iops = {
#if !defined(KERNEL_2_3)
    default_file_ops:   &lis_streams_fops,
#endif
};

/*
 * File system operations
 */
extern struct super_block *lis_read_superblock(struct super_block *sb,
					       void *ptr,
					       int silent) ;
struct file_system_type
lis_file_system_ops =
{
    name:	"LiS",
    read_super:	lis_read_superblock,
#if defined(KERNEL_2_3)
    owner:	THIS_MODULE,
    fs_flags:	(FS_NOMOUNT | FS_SINGLE)
#endif
} ;
#define	LIS_SB_MAGIC	( ('L' << 16) | ('i' << 8) | 'S' )

struct vfsmount		*lis_mnt ;

/*
 * Super block operations
 */
#if defined(KERNEL_2_3)
int lis_statfs(struct super_block *sb, struct statfs *stat) ;
#else
int lis_statfs(struct super_block *sb, struct statfs *stat, int i) ;
#endif
extern void lis_put_inode_fcn(struct inode *) ;
struct super_operations lis_super_ops =
{
    put_inode:		lis_put_inode_fcn,
    statfs:		lis_statfs
} ;


/* some kernel memory has been free'd 
 * tell STREAMS
 */
void lis_memfree( void )
{
}/*lis_memfree*/

/* This will copyin usr string pointed by ustr and return the result  in
 * *kstr. It will stop at  '\0' or max bytes copyed in.
 * caller should call FREE(*kstr) on success.
 * Will return 0 or errno
 * STATUS: complete, untested
 */
int 
lis_copyin_str(struct file *f, const char *ustr, char **kstr, int max)
{
	int    error;
#if !defined(KERNEL_2_1)
	char  *tmp, c;
	int    i=max;
#endif
	char  *mem ;

	(void) f ;
	if (max <= 0)
	    return(-ENOMEM);
	error = -EFAULT;
#if !defined(KERNEL_2_1)
	c = get_fs_byte(ustr++);
	if (!c)
	    return(-EINVAL);
#endif

	if((mem = ALLOCF(max,"copyin-buf ")) == NULL)
	    return(-ENOMEM);

#if defined(KERNEL_2_1)
	*kstr = mem ;
	error = strncpy_from_user(mem, ustr, max) ;
	if (error < 0)
	{
	    FREE(mem);
	    return(error) ;
	}

	if (error >= max)
	{
	    FREE(mem);
	    return(-ENAMETOOLONG) ;
	}

	return(0) ;
#else
	*kstr = tmp = mem ;
	while (--i)
	{
	    *(tmp++) = c;
	    c = get_fs_byte(ustr++);
	    if (!c) {
		*tmp = '\0';
		return(0);
	    }
	}
	FREE(mem);
	return(error);
#endif
}/*lis_copyin_str*/


/*  -------------------------------------------------------------------  */
/* get a err msg for error level
 */
char  *lis_errmsg( int lvl )
{
    switch(lvl){
    case LIS_PANIC:
	return(KERN_EMERG);
    case LIS_ERROR:
	return(KERN_ERR);
    case LIS_WARN:
	return(KERN_WARNING);
    case LIS_DEBUG:
	return(KERN_DEBUG);
    default:
	return("unknown severity");
    }
} /*lis_errmsg*/

/*  -------------------------------------------------------------------  */
/*				    Timeouts                             */

void
lis_tmout(struct timer_list *tl, void (*fn)(ulong), long arg, long ticks)
{
    init_timer(tl);
    tl->function= fn;
    tl->data	= arg;
    tl->expires	= jiffies + ticks;
    add_timer(tl);
}

void
lis_untmout( struct timer_list *tl)
{
    del_timer(tl);
}

/************************************************************************
*                             lis_time_till                             *
*************************************************************************
*									*
* Given a target time in terms of elapsed milli-seconds, in other words,*
* the same units as jiffies if jiffies were in milli-seconds, return	*
* the number of milli-seconds that it will take to reach to target time.*
*									*
* The return is negative if the target time is in the past, zero	*
* if it is the same as the current target time, and positive if		*
* target time is in the future.						*
*									*
************************************************************************/
long	lis_time_till(long target_time)
{
    return( target_time - jiffies*(1000/HZ) ) ;

} /* lis_time_till */

/************************************************************************
*                           lis_target_time                             *
*************************************************************************
*									*
* Convert the milli_sec interval to an absolute target time expressed	*
* in milli-seconds.  We compute this relative to the Linux software	*
* clock, jiffies, converted to milliseconds.				*
*									*
************************************************************************/
long	lis_target_time(long milli_sec)
{
    return( jiffies*(1000/HZ) + milli_sec ) ;

} /* lis_target_time */

/************************************************************************
*                           lis_milli_to_ticks                          *
*************************************************************************
*									*
* Convert milli-seconds to native ticks suitable for use with system	*
* timeout routines.							*
*									*
************************************************************************/
long	lis_milli_to_ticks(long milli_sec)
{
    return(milli_sec/(1000/HZ)) ;
}


/*  -------------------------------------------------------------------  */

/* find the head for a file descriptor.
 * STATUS: complete, untested
 */
stdata_t *
lis_fd2str( int fd )
{
#if defined(KERNEL_2_1)			/* kernel 2.1.0 or later */

    struct file		* file;
    stdata_t		* hd ;

    if ( (file = fcheck(fd)) == NULL)
	return(NULL) ;

    hd = FILE_STR(file) ;
    if (hd != NULL && hd->magic == STDATA_MAGIC)
	return(hd) ;			/* looks like a stdata structure */

    return(NULL) ;			/* not a streams file */

#else					/* kernel 2.0.xx  */
    struct file		* file;
    stdata_t		* hd ;

    if (fd>=NR_OPEN || !(file=current->files->fd[fd]))
	    return(NULL);

    hd = FILE_STR(file);
    if (hd != NULL && hd->magic == STDATA_MAGIC)
	return(hd) ;			/* looks like a stdata structure */

    return(NULL) ;			/* not a streams file */
#endif
}

struct inode *lis_file_inode(struct file *f)
{
#if defined(KERNEL_2_1)
    return((f)->f_dentry->d_inode) ;
#else
    return((f)->f_inode) ;
#endif
}

char *lis_file_name(struct file *f)
{
#if defined(KERNEL_2_1)
    return((char *)((f)->f_dentry->d_name.name)) ;
#else
    return((char *)((f)->f_inode->i_name)) ; /* FIXME */
#endif
}

struct stdata *lis_file_str(struct file *f)
{
    return((struct stdata *)(f)->private_data) ;
}

void lis_set_file_str(struct file *f, struct stdata *s)
{
    f->private_data = (void *) s ;
}

struct stdata *lis_inode_str(struct inode *i)
{
#if defined(KERNEL_2_3)
    return((struct stdata *)(i)->u.generic_ip) ;
#else
    /*
     * for 2.1.x, we need to use inode.u as pipe_i for FIFOs and pipes, but
     * we don't need base, since we don't use it for message buffers.  So,
     * we have pipe_i.base available.  (I hate doing this...)
     */
    return((struct stdata *)(i)->u.pipe_i.base) ;
#endif
}

void lis_set_inode_str(struct inode *i, struct stdata *s)
{
#if defined(KERNEL_2_3)
    (i)->u.generic_ip = (void *) s ;
#else
    (i)->u.pipe_i.base = (void *) s ;
#endif
}

struct dentry *lis_d_alloc_root(struct inode *i)
{
#if defined(KERNEL_2_3)
    return(d_alloc_root(i)) ;
#else  /* KERNEL_2_1 */
    return(d_alloc_root((i),NULL)) ;
#endif
}

void	lis_dput(struct dentry *d)
{
    dput(d) ;
}

struct dentry *lis_dget(struct dentry *d)
{
#if 0
    /* This is harmful.  It marks the dentry for, say, /dev/clone as having
     * LiS d_ops.  But if LiS is unloaded the dentry remains, pointing to
     * unmapped memory.  Since we now allocate our own inodes and dentries
     * for streams we don't need this anymore.  We never operate with the
     * kernel-assigned dentry/inode that open is called with, which is what
     * this routine is called to get.
     */
    if (d->d_op == NULL)
	d->d_op = &lis_dentry_ops;
#endif
    return(dget(d)) ;
}

/************************************************************************
*                          lis_select                                   *
*************************************************************************
*									*
* Called from the Linux sys_select routine to interrogate the status	*
* of a stream.								*
*									*
* Return 1 if the stream is ready for the requested operation.  	*
* Return 0, and set up the wait queue entry, if it is not.		*
*									*
************************************************************************/
#if !defined(LINUX_POLL)
int      lis_select(struct inode *inode, struct file *file,
		    int sel_type, select_table *wait)
{
    stdata_t		*hd;
    polldat_t		 pdat ;
    int			 evts ;
    int			 msk ;
    int			 rtn ;

    memset(&pdat, 0, sizeof(pdat)) ;		/* just in case */
retry:
    switch (sel_type)
    {
    case SEL_IN:
	pdat.pd_events = POLLIN | POLLPRI | POLLRDNORM | POLLRDBAND ;
	msk = POLLNVAL ;
	break ;
    case SEL_OUT:
	pdat.pd_events = POLLOUT | POLLWRNORM | POLLWRBAND ;
	msk = POLLNVAL ;
	break ;
    case SEL_EX:
    default:
	pdat.pd_events = POLLMSG ;			/* SIG-POLL msg */
	msk = POLLERR | POLLHUP | POLLNVAL ;
	break ;
    }

    evts = lis_strpoll(inode, file, &pdat) ;	/* get stream status */
    rtn = ( (evts & (pdat.pd_events | msk)) != 0 ) ;/* expected status */
    if (rtn) return(1) ;			/* yes, can do that */

    /*
     * Requested events are not present, add ourselves to the
     * wait queue for the stream.  Set a flag to be noticed by
     * str_rput.
     *
     * If we get back from select_wait and the flag has been unset then
     * lis_select_wakeup got called before returning to us and possibly
     * before we even got into the wait list.  In that case, we go around
     * again and return the actual events.
     */
    hd = FILE_STR(file) ;
    F_SET(hd->sd_flag,STRSELPND);
    select_wait(&hd->sd_select.sel_wait, wait) ;
    if (!F_ISSET(hd->sd_flag,STRSELPND))	/* lost the race */
	goto retry ;

    return(0) ;

} /* lis_select */
#endif

/************************************************************************
*                        lis_select_wakeup                              *
*************************************************************************
*									*
* Wake up those waiting on this stream.					*
*									*
************************************************************************/
#if !defined(LINUX_POLL)
void lis_select_wakeup(stdata_t *hd)
{
    F_CLR(hd->sd_flag,STRSELPND);	/* no longer waiting on select */
    if (   hd == NULL
	|| hd->magic != STDATA_MAGIC
	|| hd->sd_select.sel_wait == NULL
	|| hd->sd_select.sel_wait == WAIT_QUEUE_HEAD(&hd->sd_select.sel_wait)
       )
	return ;

    wake_up_interruptible(&hd->sd_select.sel_wait) ;

} /* lis_select_wakeup */
#endif

#if defined(KERNEL_2_1)
/*
 *  fattach/fdetach support
 */

#if defined(KERNEL_2_3)
static LIST_HEAD(fattaches);
#endif

#if defined(KERNEL_2_1)
#ifndef check_parent
#define check_parent(parent,dentry) \
        ((parent) == (dentry)->d_parent && !list_empty(&dentry->d_hash))
#endif
#endif

void lis_show_inode_aliases(struct inode *);  /* below */

/*
 *  d_mounted() - TRUE if the dentry's STREAM is mounted
 */
static int d_mounted( struct dentry *dm )
{
#if defined(KERNEL_2_3)
    struct list_head *e = dm->d_vfsmnt.next;
    struct vfsmount *mnt;
    struct dentry *d_covers;

    if (list_empty(&(dm->d_vfsmnt)))  return 0;

    mnt = list_entry( e, struct vfsmount, mnt_clash );
    d_covers = mnt->mnt_mountpoint;
    return (dm != d_covers);
#else
    return (dm != dm->d_covers);
#endif
}

/*
 *  d_covered() - TRUE if the dentry's path is mounted by a STREAM
 */
static int d_covered( struct dentry *dc )
{
#if defined(KERNEL_2_3)
    struct list_head *e = dc->d_vfsmnt.next;
    struct vfsmount *mnt;
    struct dentry *d_mounts;

    if (list_empty(&(dc->d_vfsmnt)))  return 0;

    mnt = list_entry( e, struct vfsmount, mnt_clash );
    d_mounts = mnt->mnt_root;
    return (dc != d_mounts);
#else
    return (dc != dc->d_mounts);
#endif
}

/*
 *  d_mounts() - return the mounted dentry
 */
static struct dentry *d_mounts( struct dentry *dc )
{
#if defined(KERNEL_2_3)
    struct list_head *e = dc->d_vfsmnt.next;
    struct vfsmount *mnt;

    if (list_empty(&(dc->d_vfsmnt)))  return dc;

    mnt = list_entry( e, struct vfsmount, mnt_clash );
    return mnt->mnt_root;
#else
    return dc->d_mounts;
#endif
}

/*
 *  d_covers() - returns the covered dentry
 */
static struct dentry *d_covers( struct dentry *dm )
{
#if defined(KERNEL_2_3)
    struct list_head *e = dm->d_vfsmnt.next;
    struct vfsmount *mnt;

    if (list_empty(&(dm->d_vfsmnt)))  return dm;

    mnt = list_entry( e, struct vfsmount, mnt_clash );
    return mnt->mnt_mountpoint;
#else
    return dm->d_covers;
#endif
}

/*
 *  dentry_mount() - mount dm on dc, using dc's parent mnt struct
 */
static int dentry_mount( struct dentry *dc, struct dentry *dm,
			 struct vfsmount *parent )
{
#if defined(KERNEL_2_3)
    /*  note that for >2.3.x, we use a pair of vfsmount structures,
     *  in order to be able to keep a reference from both the covered
     *  path and the mounted STREAM.  These are allocated as a pair, then
     *  used as if separate entries (like r/w queues).  They are kept
     *  on the fattaches list only by the first entry.
     */
    struct vfsmount *mm = kmalloc( 2*sizeof(struct vfsmount), GFP_KERNEL );
    struct vfsmount *mc = mm++;

    if (!mc)  return -ENOMEM;

    if (LIS_DEBUG_FATTACH) {
	printk("dentry_mount( 0x%x, 0x%x ) before:\n", (int)dc, (int)dm );
	lis_show_inode_aliases(dc->d_inode);
	lis_show_inode_aliases(dm->d_inode);
    }

    memset( mc, 0, sizeof(struct vfsmount) );
    atomic_set( &(mc->mnt_count), 1 );
    mc->mnt_mountpoint = dget(dc);
    mc->mnt_root = dget(dm);
    mc->mnt_parent = parent;
    list_add(&mc->mnt_clash, &dc->d_vfsmnt);
    INIT_LIST_HEAD(&mc->mnt_mounts);
    INIT_LIST_HEAD(&mc->mnt_child);
    INIT_LIST_HEAD(&mc->mnt_instances);
    list_add(&mc->mnt_list, &fattaches);

    memset( mm, 0, sizeof(struct vfsmount) );
    atomic_set( &(mm->mnt_count), 1 );
    mm->mnt_mountpoint = dget(dc);
    mm->mnt_root = dget(dm);
    mm->mnt_parent = parent;
    list_add(&mm->mnt_clash, &dm->d_vfsmnt);
    INIT_LIST_HEAD(&mm->mnt_mounts);
    INIT_LIST_HEAD(&mm->mnt_child);
    INIT_LIST_HEAD(&mm->mnt_instances);
    INIT_LIST_HEAD(&mm->mnt_list);

    if (LIS_DEBUG_FATTACH) {
	printk("dentry_mount( 0x%x, 0x%x ) after:\n", (int)dc, (int)dm );
	lis_show_inode_aliases(dc->d_inode);
	lis_show_inode_aliases(dm->d_inode);
	printk("dentry_mount() done\n" );
    }

    return 0;
#else
    if (LIS_DEBUG_FATTACH) {
	printk("dentry_mount( 0x%x, 0x%x ) before:\n", (int)dc, (int)dm );
	lis_show_inode_aliases(dc->d_inode);
	lis_show_inode_aliases(dm->d_inode);
    }

    dc->d_mounts = dget(dm);
    dm->d_covers = dget(dc);

    if (LIS_DEBUG_FATTACH) {
	printk("dentry_mount( 0x%x, 0x%x ) after:\n", (int)dc, (int)dm );
	lis_show_inode_aliases(dc->d_inode);
	lis_show_inode_aliases(dm->d_inode);
	printk("dentry_mount() done\n" );
    }

    return 0;
#endif
}

static int dentry_unmount( struct dentry *dc, struct dentry *dm )
{
#if defined(KERNEL_2_3)
    struct list_head *e;
    struct vfsmount *mnt;

    if (LIS_DEBUG_FATTACH) {
	printk("dentry_unmount( 0x%x, 0x%x ) before:\n", (int)dc, (int)dm );
	lis_show_inode_aliases(dc->d_inode);
	lis_show_inode_aliases(dm->d_inode);
    }

    if (!list_empty(&(dc->d_vfsmnt))) {
	e = dc->d_vfsmnt.next;
	mnt = list_entry( e, struct vfsmount, mnt_clash );
	list_del(&mnt->mnt_list);
	kfree(mnt);
    }

    INIT_LIST_HEAD(&(dc->d_vfsmnt));
    INIT_LIST_HEAD(&(dm->d_vfsmnt));

    if (LIS_DEBUG_FATTACH) {
	printk("dentry_unmount( 0x%x, 0x%x ) after:\n", (int)dc, (int)dm );
	lis_show_inode_aliases(dc->d_inode);
	lis_show_inode_aliases(dm->d_inode);
	printk("dentry_unmount() done\n" );
    }

    /*
     *  we did 2 dget()'s in dentry_mount() for each of these - undo them now
     */
    dput(dc);  dput(dc);
    dput(dm);  dput(dm);

    return 0;
#else
    if (LIS_DEBUG_FATTACH) {
	printk("dentry_unmount( 0x%x, 0x%x ) before:\n", (int)dc, (int)dm );
	lis_show_inode_aliases(dc->d_inode);
	lis_show_inode_aliases(dm->d_inode);
    }

    dc->d_mounts = dc;
    dm->d_covers = dm;

    if (LIS_DEBUG_FATTACH) {
	printk("dentry_unmount( 0x%x, 0x%x ) after:\n", (int)dc, (int)dm );
	lis_show_inode_aliases(dc->d_inode);
	lis_show_inode_aliases(dm->d_inode);
	printk("dentry_unmount() done\n" );
    }

    dput(dc);
    dput(dm);

    return 0;
#endif
}

#endif  /* KERNEL_2_1 */

/*
 *  a debugging routine to show the aliases for an inode
 */

void lis_show_inode_aliases( struct inode *i )
{
#if defined(KERNEL_2_1)
    struct list_head *ent;

    if (!LIS_DEBUG_ADDRS || list_empty(&(i->i_dentry)))  return;

    printk("inode_alias(0x%x): i_count %d i_nlink %d i_ino %d\n",
	   (int)i, I_COUNT(i), (int)i->i_nlink, (int)i->i_ino );

    for (ent = i->i_dentry.next;  ent != &(i->i_dentry);  ent = ent->next) {
	struct dentry *d = list_entry( ent, struct dentry, d_alias );

	printk("inode_alias(0x%x): dentry 0x%x(%d) name \"%s\"",
	       (int)i, (int)d, d->d_count,
	       d->d_name.name );
	if (d_covered(d))
	    printk(" mounts 0x%x(%d) \"%s\"",
		   (int)d_mounts(d), d_mounts(d)->d_count,
		   d_mounts(d)->d_name.name );
	if (d_mounted(d))
	    printk(" covers 0x%x(%d) \"%s\"",
		   (int)d_covers(d), d_covers(d)->d_count,
		   d_covers(d)->d_name.name );

	printk("\n");
    }
#endif
}

/************************************************************************
*                           lis_statfs                                  *
*************************************************************************
*									*
* Return file system stats.						*
*									*
************************************************************************/
#if defined(KERNEL_2_3)
int lis_statfs(struct super_block *sb, struct statfs *stat)
#else
int lis_statfs(struct super_block *sb, struct statfs *stat, int i)
#endif
{
    stat->f_type = LIS_SB_MAGIC ;
    stat->f_bsize = 1024 ;
    stat->f_namelen = 255 ;
    return(0) ;

}

/************************************************************************
*                         lis_read_superblock                           *
*************************************************************************
*									*
* The 2.4 kernel requires all file systems to maintain at least the	*
* facade of having a superblock.  We do that here, filling in dummy	*
* values in the structures.						*
*									*
************************************************************************/
struct super_block *lis_read_superblock(struct super_block *sb,
				        void		   *ptr,
				        int		    silent)
{
    struct inode	*isb ;

#if defined(KERNEL_2_3)			/* 2.4 kernel */
    isb = new_inode(sb) ;
#else					/* 2.2 kernel */
    isb = get_empty_inode() ;
#endif
    if (isb == NULL)
	return(NULL) ;

    isb->i_mode  = S_IFDIR | S_IRUSR | S_IWUSR ;
    isb->i_uid   = 0 ;
    isb->i_gid   = 0 ;
    isb->i_atime = CURRENT_TIME ;
    isb->i_mtime = CURRENT_TIME ;
    isb->i_ctime = CURRENT_TIME ;

    sb->s_magic		= LIS_SB_MAGIC ;
    sb->s_blocksize	= 1024 ;
    sb->s_blocksize_bits= 10 ;
    sb->s_op		= &lis_super_ops ;
    sb->s_root		= d_alloc(NULL, &(const struct qstr) {"LiS:", 4, 0});
    if (sb->s_root == NULL)
    {
	iput(isb) ;
	return(NULL) ;
    }

    sb->s_root->d_sb     = sb;
    sb->s_root->d_parent = sb->s_root;
    d_instantiate(sb->s_root, isb);
    return(sb) ;
}

/************************************************************************
*                         lis_delete_dentry                             *
*************************************************************************
*									*
* This routine has to be here (in 2.4 kernel) and has to return 1 in	*
* order to get the kernel's dput routine to go ahead and iput the inode	*
* beneath the dentry when the dentry's d_count goes to zero.  Duh.  Like*
* the default behavior should be to just keep the dangling inode around	*
* until the file system unmounts so that then the kernel can print a	*
* message about dangling inodes.					*
*									*
************************************************************************/
#if defined(KERNEL_2_3)
int lis_delete_dentry(struct dentry *dentry)
{
    return(1) ;
}
#elif defined(KERNEL_2_1)
void lis_delete_dentry(struct dentry *dentry)
{
}
#endif

/************************************************************************
*                        lis_put_inode_fcn                              *
*************************************************************************
*									*
* Called when an LiS inode is being "iput".				*
*									*
************************************************************************/
void lis_put_inode_fcn(struct inode *i)
{
#if defined(KERNEL_2_3)			/* linux >= 2.3.0 */
    lis_atomic_dec(&lis_inode_cnt) ;
#endif
    (void) i ;
}


/************************************************************************
*                            lis_new_inode                              *
*************************************************************************
*									*
* This routine is called from stropen().				*
*									*
* 'old' is the inode passed to stropen for a clone open.  'oldf'	*
* is the file structure for the file that was opened.  The stdata	*
* structure hangs off of the file structure.				*
*									*
* 'dev' is the major/minor for the device that was really opened.	*
* Presumably, we are called because the i_rdev of the inode struct	*
* does not match up with dev.						*
*									*
* We need to allocate a new inode struct, and dentry for 2.2 kernels,	*
* and get them linked to this file.  The old inode and dentry		*
* represent the clone device and not the device that was actually	*
* opened.								*
*									*
* This operation is performed for every stream open whether clone or not*
*									*
************************************************************************/
struct inode *
lis_new_inode( struct file *f, dev_t dev )
{
    struct inode *old = FILE_INODE(f);
    struct inode *new;
    stdata_t	 *hd = FILE_STR(f) ;
#ifdef KERNEL_2_1
    struct dentry *oldd = f->f_dentry;
#endif
#ifdef KERNEL_2_3
    struct qstr    dname ;
#endif

    if (LIS_DEBUG_OPEN && 0)
	printk("new_inode( 0x%x, 0x%x ) "
#ifdef KERNEL_2_1
	       "old dentry 0x%x(%d) "
#endif	       
	       "old inode 0x%x(%d)\n",
	       (int)f, (int)dev,
#ifdef KERNEL_2_1
	       (int)oldd, oldd->d_count,
#endif
	       (int)old, I_COUNT(old) );

    if (!old)  {				/* param checking */
	printk("lis_new_inode: old inode must be non-NULL\n") ;
	return(NULL) ;				/* bad return */
    }

    if (hd == NULL)
    {
	printk("lis_new_inode: file must have stream head ptr\n") ;
	return(NULL) ;				/* bad return */
    }

    /*
     *  we keep an old inode only if it's a LiS-only inode.  Otherwise,
     *  we'll replace a non-LiS inode here with one that is LiS-only
     */
    if (major(old->i_dev) == lis_major && old->i_rdev == dev) {
	return (old);
    }

    if ((new = get_empty_inode()))  {		/* got a new inode */
	new->i_op    = &lis_streams_iops;
#ifdef KERNEL_2_3
	new->i_fop   = &lis_streams_fops;
#endif
	new->i_state = I_DIRTY;       /* keep it off the dirty list */
	/*
	 *  char devs are identified by the rdev field; dev identifies
	 *  the hosting file system.  Here, we construct our own dev
	 *  field, reflecting that this is a LiS-only inode which has
	 *  no file system hosting it (other than LiS itself)
	 */
	new->i_rdev  = dev;			/* set desired dev */
	new->i_dev   = makedevice( lis_major, major(dev) );
	/**/
	new->i_mode  = old->i_mode;		/* inherit mode */
	new->i_ctime = old->i_ctime;		/* inherit times */
	new->i_atime = old->i_atime;
	new->i_mtime = old->i_mtime;
	new->i_uid   = old->i_uid;		/* inherit owner */
	new->i_gid   = old->i_gid;
#if defined(KERNEL_2_1) && !defined(KERNEL_2_3)
	/*
	 *  this stuff is only used for FIFOs/pipes, but it's statically
	 *  part of the inode, and we don't know yet if this is a
	 *  FIFO/pipe inode.  It doesn't hurt to do this in any event...
	 *
	 *  note also that for 2.1.x/2.2.x, we are using PIPE_BASE as
	 *  INODE_STR.  Since this is a new inode, we're OK here setting
	 *  it to NULL.
	 */
	PIPE_BASE(*new) = NULL;
	PIPE_WAIT(*new) = NULL;
	PIPE_START(*new) = PIPE_LEN(*new) = 0;
	PIPE_RD_OPENERS(*new) = PIPE_WR_OPENERS(*new) = 0;
	PIPE_READERS(*new) = PIPE_WRITERS(*new) = 0;
	PIPE_LOCK(*new) = 0;
#endif
#if defined(KERNEL_2_1)				/* linux >= 2.1.0 */
 #if defined(KERNEL_2_3)			/* linux >= 2.3.0 */
	new->i_sb   = lis_mnt->mnt_sb ;		/* superblock ptr */
	lis_atomic_inc(&lis_inode_cnt) ;
	dname.name = hd->sd_name ;
	dname.len  = strlen(hd->sd_name) ;
	dname.hash = new->i_ino;
	f->f_dentry = d_alloc(lis_mnt->mnt_sb->s_root, &dname);
	if (f->f_dentry == NULL)
	{
	    iput(new) ;
	    return(NULL) ;
	}

	f->f_dentry->d_op = &lis_dentry_ops;
	d_add(f->f_dentry, new);
 #else						/* 2.2 kernel */
	f->f_dentry = lis_d_alloc_root(new) ;
	if (f->f_dentry == NULL)
	{
	    iput(new) ;
	    return(NULL) ;
	}

	f->f_dentry->d_op = &lis_dentry_ops;
 #endif
	dput(oldd);     			/* free old inode/dentry */
#else						/* linux <= 2.0.xx */
	FILE_INODE(f) = new;    		/* hook file to new inode */
	iput(old);				/* free old inode */
#endif
    }

    if (LIS_DEBUG_OPEN && 0)
	printk("new_inode( 0x%x, 0x%x ) "
#ifdef KERNEL_2_1
	       "new dentry 0x%x(%d) "
#endif	       
	       "new inode 0x%x(%d)\n",
	       (int)f, (int)dev,
#ifdef KERNEL_2_1
	       (int)f->f_dentry, f->f_dentry->d_count,
#endif
	       (int)new, I_COUNT(new) );

    return(new) ;				/* use new inode */
    						/* possibly NULL return */
} /* lis_new_inode */

/*
 *  lis_get_filp() - get a LiS-ready file pointer, or set one up for LiS
 */
struct file *
lis_get_filp( struct file *f )
{
    if (!f && !(f = get_empty_filp()))  return NULL;

    f->f_pos   = 0;
    f->f_op    = &lis_streams_fops;
    f->f_flags = O_RDWR;
    f->f_mode  = FMODE_READ|FMODE_WRITE;

    return f;
}

/*
 * lis_set_up_inode() - make an inode look like an LiS inode
 *
 * Used in stream head when a stream file is opened.
 */
struct inode *lis_set_up_inode(struct file *f, struct inode *inode)
{
    struct inode	*new ;

    if (inode == NULL) return(NULL) ;

    new = lis_new_inode(f, inode->i_rdev) ;
    if (new == NULL)
	return(NULL) ;

    /*
     * Set the user/group ids to the opener, set modification times
     * to the current time.
     */
    new->i_uid   = current->fsuid;
    new->i_gid   = current->fsgid;
    new->i_atime = new->i_mtime = new->i_ctime = CURRENT_TIME;
    return(new) ;
}

/*
 *  lis_get_inode() - allocate and setup a LiS-only inode
 */
static struct inode *
lis_get_inode( mode_t mode, dev_t dev )
{
    struct inode *inode = get_empty_inode();

    if (inode)
    {
	inode->i_mode  = mode;
	/*
	 *  see the comment above (in lis_new_inode) about the use of
	 *  the dev and rdev fields
	 */
	inode->i_rdev  = dev;
	inode->i_dev   = makedevice( lis_major, major(dev) );
    }

    return inode;
}

/*
 *  lis_old_inode() - swap the inode referenced by f with that referenced
 *  by i.
 */
struct inode *
lis_old_inode( struct file *f, struct inode *i )
{
#ifdef KERNEL_2_1
    struct dentry *oldd = f->f_dentry;
#else
    struct inode *oldi = f->f_inode;
#endif

    if (LIS_DEBUG_OPEN && 0) {
	printk(
	    "old_inode( 0x%x, 0x%x(%d) ) "
	    "inode 0x%x(%d) dev 0x%x -> 0x%x\n",
	    (int)f,
	    (int)i, I_COUNT(i),
	    (int)FILE_INODE(f), I_COUNT(FILE_INODE(f)),
	    (int)FILE_INODE(f)->i_rdev, (int)i->i_rdev );
    }

#ifdef KERNEL_2_1
    LOCK_INO(i);
    f->f_dentry = lis_d_alloc_root(igrab(i));
    ULOCK_INO(i);
    dput(oldd);
#else
    LOCK_INO(i);
    f->f_inode = igrab(i);
    ULOCK_INO(i);
    iput(oldi);
#endif

    if (LIS_DEBUG_OPEN && 0)
	printk(
	    "old_inode( ) inode 0x%x(%d)\n",
	    (int)FILE_INODE(f), I_COUNT(FILE_INODE(f)) );

    return (FILE_INODE(f));
}

struct inode *lis_grab_inode(struct inode *ino)
{
    return(igrab(ino)) ;
}

void lis_put_inode(struct inode *ino)
{
    iput(ino) ;
}


/*
 *  lis_get_fifo() - create a new unique fifo
 *
 *  note that lis_get_filp() uses lis_streams_fops, which is what makes
 *  mode S_IFIFO a STREAM instead of a kernel FIFO; i.e., it ensures
 *  that stropen will be called to open the FIFO.
 */
int lis_get_fifo( struct file **f )
{
    mode_t mode = S_IFCHR|S_IRUSR|S_IWUSR;
    dev_t dev = makedevice( LIS_CLONE, LIS_FIFO );
    struct inode *inode = lis_get_inode( mode, dev );
    int error;

    error = -ENFILE;
    if (!inode)  goto no_inode;

    inode->i_mode &= ~current->fs->umask ;	/* umask considerations */

    if (!(*f = lis_get_filp(NULL)))  goto no_file;

    error = -ENOMEM;
#ifdef KERNEL_2_1
    if (!((*f)->f_dentry = lis_d_alloc_root(inode)))
	goto no_stream;
#else
    (*f)->f_inode = inode;
#endif

    if ((error = lis_stropen( inode, *f )) < 0)  goto no_stream;

    return 0;

no_stream:
    fput(*f);
    *f = NULL;
no_file:
    iput(inode);
no_inode:
    return error;
}

#if 0		/* shimmed in by liskmod.c */
/*
 *  put_unused_fd() -
 *
 *  this is really defined in the kernel.  If the kernel is ever patched
 *  to make it available, the kernel's version should be used.  So that
 *  things will compile, this no-op version is included here.
 *
 *  note that it is only needed when a pipe or fifo open fails, and that
 *  should almost never happen.  Moreover, the only ill effect of not
 *  doing anything here is that an otherwise unused fd will be "wasted"
 *  in a process which could only fail because it has used too many.  So,
 *  it's really not a big deal if it's not really working.
 */
static inline void put_unsigned_fd( unsigned int fd )  {}
#endif

/*
 *  lis_get_pipe()
 *
 *  create a new unique pipe as two new unique file pointers and all
 *  lower "plumbing" from them (i.e., inodes, stream heads, queues).
 *
 *  the process is simply to create two new unique FIFOs, and "twist"
 *  them.
 */
int lis_get_pipe( struct file **f0, struct file **f1 )
{
    stdata_t *hd0, *hd1;
    queue_t *wq0, *wq1;
    int error;

    /*
     *  get a pair of unique FIFOs for the pipe ends
     */
    if ((error = lis_get_fifo( f0 )) < 0)  goto no_fifos;
    if ((error = lis_get_fifo( f1 )) < 0)  goto no_fifo1;

    /*
     *  OK - make them peers, and get the head queues for twisting
     */
    hd0 = FILE_STR(*f0);  wq0 = hd0->sd_wq;
    hd1 = FILE_STR(*f1);  wq1 = hd1->sd_wq;
    hd0->sd_peer = hd1;
    hd1->sd_peer = hd0;

    /*
     *  twist the write queues to point to the peers' read queues
     */
    wq0->q_next = RD(wq1);
    wq1->q_next = RD(wq0);

    return 0;
    
no_fifo1:
    lis_strclose( FILE_INODE(*f0), *f0 );
no_fifos:
    return error;
}

/*
 *  lis_pipe()
 *
 *  this routine is just an intermediate layer in the pipe system call
 *  process.  It's job is to turn the file pointers returned by lis_get_pipe()
 *  into a pair of file descriptors.
 */
int
lis_pipe( unsigned int *fd )
{
    int fd0, fd1;
    struct file *fp0, *fp1;
    int error;

    if ((fd0 = get_unused_fd()) < 0) {
	error = fd0;  goto no_fds;
    }
    if ((fd1 = get_unused_fd()) < 0) {
	error = fd1;  goto no_fd1;
    }

    if ((error = lis_get_pipe( &fp0, &fp1 )) < 0)  goto no_pipe;

    /*
     *  got everything - hook it up
     */
    fd_install( fd0, fp0 );
    fd_install( fd1, fp1 );
    fd[0] = fd0;
    fd[1] = fd1;

    return 0;
    
no_pipe:
    put_unused_fd(fd1);
no_fd1:
    put_unused_fd(fd0);
no_fds:
    return error;
}

/*
 *  pipe() ioctl wrapper (pretty much the same as the syscall)
 *	   syscall wrapper is in syscalls.c
 */
int
lis_ioc_pipe( unsigned int *fildes )
{
    unsigned int fd[2];
    int error;
    
    error = lis_pipe(fd);
    if (!error) {
	if (copy_to_user(fildes, fd, 2*sizeof(unsigned int)))
	    error = -EFAULT;
    }
    return error;
}

/*
 *  the following ..._sync() routines are adaptations of the kernel
 *  fifo open routines, used here to implement conventional FIFO
 *  synchronization at open/close time (and elsewhere).  We actually
 *  use the same data structures, so it makes sense to model this
 *  code closely after the kernel's code.  Note that this is only
 *  done for FIFOs and pipes, and only actually needed for O_RDONLY
 *  and O_WRONLY for FIFOs, but it's here for sake of completeness...
 *
 *  Note also that this is safe to do.  LiS now uses inodes which
 *  are totally independent of any host filesystem - we can do whatever
 *  we want with them (as long as the kernel doesn't complain).
 *
 *  the preliminary routines here are support for fifo_open_sync.
 */

#if defined(KERNEL_2_3)
static inline void lis_fifo_wait(struct inode * i)
{
	DECLARE_WAITQUEUE(wait, current);
	current->state = TASK_INTERRUPTIBLE;
	add_wait_queue(PIPE_WAIT(*i), &wait);
	lis_kernel_up(PIPE_SEM(*i));
	schedule();
	remove_wait_queue(PIPE_WAIT(*i), &wait);
	current->state = TASK_RUNNING;
	lis_kernel_down(PIPE_SEM(*i));
}

static inline void lis_fifo_wait_for_partner( struct inode* i,
					      unsigned int* cnt)
{
    int cur = *cnt;	
    while (cur == *cnt) {
	lis_fifo_wait(i);
	if (signal_pending(current))
	    break;
    }
}

static inline void lis_fifo_wake_up_partner(struct inode* i)
{
    lis_wake_up_interruptible(PIPE_WAIT(*i));
}

static struct inode* lis_fifo_info_new(struct inode* i)
{
    i->i_pipe = kmalloc(sizeof(struct pipe_inode_info), GFP_KERNEL);
    if (i->i_pipe) {
	init_waitqueue_head(PIPE_WAIT(*i));
	PIPE_BASE(*i) = NULL;;
	PIPE_START(*i) = PIPE_LEN(*i) = 0;
	PIPE_READERS(*i) = PIPE_WRITERS(*i) = 0;
	PIPE_WAITING_READERS(*i) = PIPE_WAITING_WRITERS(*i) = 0;
	PIPE_RCOUNTER(*i) = PIPE_WCOUNTER(*i) = 1;
	
	return i;
    } else {
	return NULL;
    }
}
#endif /* KERNEL_2_3 */

int
lis_fifo_open_sync( struct inode *i, struct file *f )
{
#if defined(KERNEL_2_1)
    stdata_t *head = INODE_STR(i);
    long this_open = lis_open_cnt;
    int ret = 0;

    if (LIS_DEBUG_OPEN)
	printk("fifo_open_sync#%ld: "
	       "%s%s file mode 0%o flags 0%o\n",
	       this_open,
	       lis_strm_name(head), lis_maj_min_name(head),
	       (int)f->f_mode, (int)f->f_flags );

#if defined(KERNEL_2_3)

    ret = -ERESTARTSYS;
    if (lis_kernel_down(PIPE_SEM(*i)))
	goto err_nolock_nocleanup;
    
    if (!i->i_pipe) {
	ret = -ENOMEM;
	if(!lis_fifo_info_new(i))
	    goto err_nocleanup;
    }
    f->f_version = 0;

    switch (f->f_mode) {
    case 1:
	/*
	 *  O_RDONLY
	 *  POSIX.1 says that O_NONBLOCK means return with the FIFO
	 *  opened, even when there is no process writing the FIFO.
	 */
	PIPE_RCOUNTER(*i)++;
	if (PIPE_READERS(*i)++ == 0)
	    lis_fifo_wake_up_partner(i);
	
	if (!PIPE_WRITERS(*i)) {
	    if ((f->f_flags & O_NONBLOCK)) {
		/* suppress POLLHUP until we have seen a writer */
		f->f_version = PIPE_WCOUNTER(*i);
	    } else 
	    {
		lis_fifo_wait_for_partner(i, &PIPE_WCOUNTER(*i));
		if (signal_pending(current))
		    goto err_rd;
	    }
	}
	break;
	
    case 2:
	/*
	 *  O_WRONLY
	 *  POSIX.1 says that O_NONBLOCK means return -1 with
	 *  errno=ENXIO when there is no process reading the FIFO.
	 */
	ret = -ENXIO;
	if ((f->f_flags & O_NONBLOCK) && !PIPE_READERS(*i))
	    goto err;
	
	PIPE_WCOUNTER(*i)++;
	if (!PIPE_WRITERS(*i)++)
	    lis_fifo_wake_up_partner(i);
	
	if (!PIPE_READERS(*i)) {
	    lis_fifo_wait_for_partner(i, &PIPE_RCOUNTER(*i));
	    if (signal_pending(current))
		goto err_wr;
	}
	break;
	
    case 3:
	/*
	 *  O_RDWR
	 *  POSIX.1 leaves this case "undefined" when O_NONBLOCK is set.
	 *  This implementation will NEVER block on a O_RDWR open, since
	 *  the process can at least talk to itself.
	 */
	PIPE_READERS(*i)++;
	PIPE_WRITERS(*i)++;
	PIPE_RCOUNTER(*i)++;
	PIPE_WCOUNTER(*i)++;
	if (PIPE_READERS(*i) == 1 || PIPE_WRITERS(*i) == 1)
	    lis_fifo_wake_up_partner(i);
	break;
	
    default:
	ret = -EINVAL;
	goto err;
    }

    lis_kernel_up(PIPE_SEM(*i));

    if (LIS_DEBUG_OPEN)
	printk("fifo_open_sync#%ld: "
	       "%s%s %d reader(s) %d writer(s)\n",
	       this_open,
	       lis_strm_name(head), lis_maj_min_name(head),
	       PIPE_READERS(*i), PIPE_WRITERS(*i) );

    return 0;
    
err_rd:
    if (!--PIPE_READERS(*i))
	lis_wake_up_interruptible(PIPE_WAIT(*i));
    ret = -ERESTARTSYS;
    goto err;
    
err_wr:
    if (!--PIPE_WRITERS(*i))
	lis_wake_up_interruptible(PIPE_WAIT(*i));
    ret = -ERESTARTSYS;
    goto err;
    
err:
    if (!PIPE_READERS(*i) && !PIPE_WRITERS(*i)) {
	kfree(i->i_pipe);
	i->i_pipe = NULL;
    }

err_nocleanup:
    lis_kernel_up(PIPE_SEM(*i));

err_nolock_nocleanup:
    if (LIS_DEBUG_OPEN)
	printk("fifo_open_sync#%ld: %s%s error %d\n",
	       this_open,
	       lis_strm_name(head), lis_maj_min_name(head), ret );

    return ret;
#else    /* !KERNEL_2_3 */

    switch (f->f_mode) {
    case 1:
	/*
	 *  O_RDONLY
	 *  POSIX.1 says that O_NONBLOCK means return with the FIFO
	 *  opened, even when there is no process writing the FIFO.
	 */
	if (!PIPE_READERS(*i)++)
	    lis_wake_up_interruptible(&PIPE_WAIT(*i));
	if (!(f->f_flags & O_NONBLOCK) && !PIPE_WRITERS(*i)) {
	    PIPE_RD_OPENERS(*i)++;
	    while (!PIPE_WRITERS(*i)) {
		if (signal_pending(current)) {
		    ret = -ERESTARTSYS;
		    break;
		}
		lis_interruptible_sleep_on(&PIPE_WAIT(*i));
	    }
	    if (!--PIPE_RD_OPENERS(*i))
		lis_wake_up_interruptible(&PIPE_WAIT(*i));
	}
	while (PIPE_WR_OPENERS(*i))
	    lis_interruptible_sleep_on(&PIPE_WAIT(*i));
	if (ret && !--PIPE_READERS(*i))
	    lis_wake_up_interruptible(&PIPE_WAIT(*i));
	break;
	
    case 2:
	/*
	 *  O_WRONLY
	 *  POSIX.1 says that O_NONBLOCK means return -1 with
	 *  errno=ENXIO when there is no process reading the FIFO.
	 */
	if ((f->f_flags & O_NONBLOCK) && !PIPE_READERS(*i)) {
	    ret = -ENXIO;
	    break;
	}
	if (!PIPE_WRITERS(*i)++)
	    lis_wake_up_interruptible(&PIPE_WAIT(*i));
	if (!PIPE_READERS(*i)) {
	    PIPE_WR_OPENERS(*i)++;
	    while (!PIPE_READERS(*i)) {
		if (signal_pending(current)) {
		    ret = -ERESTARTSYS;
		    break;
		}
		lis_interruptible_sleep_on(&PIPE_WAIT(*i));
	    }
	    if (!--PIPE_WR_OPENERS(*i))
		lis_wake_up_interruptible(&PIPE_WAIT(*i));
	}
	while (PIPE_RD_OPENERS(*i))
	    lis_interruptible_sleep_on(&PIPE_WAIT(*i));
	if (ret && !--PIPE_WRITERS(*i))
	    lis_wake_up_interruptible(&PIPE_WAIT(*i));
	break;
	
    case 3:
	/*
	 *  O_RDWR
	 *  POSIX.1 leaves this case "undefined" when O_NONBLOCK is set.
	 *  This implementation will NEVER block on a O_RDWR open, since
	 *  the process can at least talk to itself.
	 */
	if (!PIPE_READERS(*i)++)
	    lis_wake_up_interruptible(&PIPE_WAIT(*i));
	while (PIPE_WR_OPENERS(*i))
	    lis_interruptible_sleep_on(&PIPE_WAIT(*i));
	if (!PIPE_WRITERS(*i)++)
	    lis_wake_up_interruptible(&PIPE_WAIT(*i));
	while (PIPE_RD_OPENERS(*i))
	    lis_interruptible_sleep_on(&PIPE_WAIT(*i));
	break;
	
    default:
	ret = -EINVAL;
    }

    if (LIS_DEBUG_OPEN)
	printk("fifo_open_sync#%ld: "
	       "%s%s %d reader(s) %d writer(s)\n",
	       this_open,
	       lis_strm_name(head), lis_maj_min_name(head),
	       PIPE_READERS(*i), PIPE_WRITERS(*i) );

    return ret;
#endif   /* !KERNEL_2_3 */
#endif   /* KERNEL_2_1 */
}

void
lis_fifo_close_sync( struct inode *i, struct file *f )
{
#if defined(KERNEL_2_1)
    stdata_t *head = INODE_STR(i);
    long this_close = lis_close_cnt;

#if defined(KERNEL_2_3)
    lis_kernel_down(PIPE_SEM(*i));
#endif

    PIPE_READERS(*i) -= (f->f_mode & FMODE_READ ? 1 : 0);
    PIPE_WRITERS(*i) -= (f->f_mode & FMODE_WRITE ? 1 : 0);

    if (LIS_DEBUG_CLOSE)
	printk("fifo_close_sync#%ld: %s%s %d reader(s) %d writer(s)\n",
	       this_close,
	       lis_strm_name(head), lis_maj_min_name(head),
	       PIPE_READERS(*i), PIPE_WRITERS(*i) );

#if defined(KERNEL_2_3)
    if (!PIPE_READERS(*i) && !PIPE_WRITERS(*i)) {
	kfree(i->i_pipe);
	i->i_pipe = NULL;
    } else {
	lis_wake_up_interruptible(PIPE_WAIT(*i));
    }

    lis_kernel_up(PIPE_SEM(*i));
#else
    lis_wake_up_interruptible(&PIPE_WAIT(*i));
#endif

#endif  /* KERNEL_2_1 */
}


int
lis_fifo_write_sync( struct inode *i, int written )
{
#if defined(KERNEL_2_1)
    if (!PIPE_READERS(*i)) {
	if (!written)
	    send_sig( SIGPIPE, current, 0 );
	return(-EPIPE);
    } else
	return 0;
#else
    return 0;
#endif
}


static void
lis_fifo_sendfd_sync( struct inode *i, struct file *f )
{
#if defined(KERNEL_2_1)
    PIPE_READERS(*i) += (f->f_mode & FMODE_READ ? 1 : 0);
    PIPE_WRITERS(*i) += (f->f_mode & FMODE_WRITE ? 1 : 0);
#endif
}

/*
 *  fattach() -
 *
 *  "mount" a stream on a non-STREAM path.  Subsequent opens of the
 *  path will open the mounted stream, until the path is fdetach()'ed.
 *
 *  Note Linux semantics:
 *  - only one mount is done, for the given path, not for all of its
 *    aliases (i.e., hard links).  The reason is that those aliases
 *    can't be efficiently loaded as dentries (if they aren't already)
 *    because the Linux FS org doesn't allow easy access to hard-linked
 *    directory entries.  It _is_ possible to link all such names that
 *    are already in the dcache, but those that are not will be missed.
 *    To have predictable semantics, then, we only mount the given path.
 *  - Permissions are kept at the underlying inodes in the Linux FS.
 *    Since a stream may be mounted on different paths (e.g., to get
 *    around the above limitation, if the user knows the aliases and
 *    calls fattach separately for each), it is not deemed reasonable
 *    to have this routine "initialize" the streams permissions & ids
 *    to those of the underlying file (since there may be many).
 *    Instead (and more simply), we just check here that the caller has
 *    read/write access to the path.
 *  - We allow mounting to any non-directory.  This allows mounting
 *    not only over regular files and streams, but also over Linux FIFOs
 *    and pipes.  This thus covers all expected usage, but also completely
 *    avoids conflict with normal Linux mount() usage, which only applies
 *    to directories.
 */
int
lis_fattach( struct file *f, const char *path )
{
#if defined(KERNEL_2_1)
    struct dentry *d_mount, *d_cover, *d_cover_p;
    struct inode *i_mount, *i_cover;
    stdata_t *head = NULL;
#ifdef KERNEL_2_3
    struct nameidata nd;
    struct vfsmount *mnt;
#else
    void *mnt = NULL;
#endif
    int error;
    int hd_locked = 0 ;

    if (LIS_DEBUG_FATTACH)
	cmn_err( CE_CONT,
		 "fattach( 0x%x, \"%s\" )\n", (int)f, path );

    /*
     *  get the inode and stream head for the file - verify the head
     */
    i_mount = f->f_dentry->d_inode;
    error = -EINVAL;
    if (major(i_mount->i_dev) != lis_major)  goto out;

    if (!(head = FILE_STR(f)) ||
	head->magic != STDATA_MAGIC ||
	head->sd_inode != i_mount)  goto out;

    lis_spin_lock(&head->sd_lock) ;		/* lock strm head */
    hd_locked = 1 ;
    if (LIS_DEBUG_FATTACH)
	cmn_err( CE_CONT,
		 "fattach( \"%s\" ) i_mount 0x%x head 0x%x(%d) %s%s\n",
		 path, (int)i_mount,
		 (int)head, LIS_REFCNT(head),
		 lis_strm_name(head), lis_maj_min_name(head) );

    /*
     *  get the dentry for the path - namei() bumps its count
     */
#ifdef KERNEL_2_3
    error = user_path_walk( path, &nd );
    if (error)  goto out;
    d_cover = nd.dentry;
    mnt = nd.mnt;
    if (LIS_DEBUG_FATTACH)
	if (mnt)
	    printk("fattach( \"%s\" ): mount 0x%x dir \"%s\" dev \"%s\"\n",
		   path, (int)nd.mnt,
		   (mnt->mnt_mountpoint ?
			(char *)mnt->mnt_mountpoint->d_name.name : "<none>"),
		   (mnt->mnt_devname ? mnt->mnt_devname : "<none>") );
    mntput(nd.mnt);
#else
    d_cover = namei(path);
    error = PTR_ERR(d_cover);
    if (IS_ERR(d_cover))  goto out;
#endif
    d_cover_p = dget(d_cover->d_parent);
    i_cover = d_cover->d_inode;

    if (LIS_DEBUG_FATTACH)
	cmn_err( CE_CONT,
		 "fattach( \"%s\" ) d_cover 0x%x inode 0x%x\n",
		 path, (int)d_cover,
		 (int)d_cover->d_inode );

    LOCK_INO(i_mount);

    /*
     *  the caller needs write permission to the inode to be covered
     */
    if ((error = permission( i_cover, MAY_WRITE )) < 0)
	goto dput_and_out;

#if defined(KERNEL_2_1)
    if (f->f_mode & FMODE_WRITE)
	if ((error = get_write_access(i_mount)) < 0)
	    goto dput_and_out;
#endif

    /*
     *  the path can't already be/have a mount
     */
    error = -EBUSY;
    if (d_mounted(d_cover) || d_covered(d_cover))
	goto dput_and_out;

    /*
     *  the path can't be a directory
     */
    error = -EISDIR;
    if (S_ISDIR(i_cover->i_mode))
	goto dput_and_out;

    /*
     *  We follow the convention that the kernel uses for locking
     *  d_cover; i.e., we lock the parent's inode.  We then check it
     *  to make sure it's still valid.
     *
     *  note that a deadlock isn't possible here; since a stream is
     *  never open as an inode which is hosted by a filesystem, d_cover
     *  will never identify a stream, and so there can be no lock
     *  cycle.
     */
    error = -ENOENT;
    LOCK_INO(d_cover_p->d_inode);
    if (!check_parent(d_cover_p,d_cover)) {
	ULOCK_INO(d_cover_p->d_inode);
	goto dput_and_out;
    }

    /*
     *  the conventional semantics for fattach() are that a stream can
     *  be mounted to multiple pathnames, and must cover all pathnames
     *  linked to the one passed in.  We support the former but not the
     *  latter.  Under Linux, we can't guarantee that the inode's alias
     *  list includes all such links.  To be predictable, then, we only
     *  mount on the given path.
     */
    d_mount = lis_d_alloc_root(igrab(i_mount));

    if (LIS_DEBUG_FATTACH)
	cmn_err( CE_CONT,
		 "fattach( \"%s\" ) d_mount 0x%x(%d) inode 0x%x(%d)\n",
		 path, (int)d_mount, d_mount->d_count,
		 (int)d_mount->d_inode, I_COUNT(d_mount->d_inode) );

    /*
     *  OK - do a mount - dentry_mount does dget() for both
     */
    if ((error = dentry_mount( d_cover, d_mount, mnt )) < 0) {
	ULOCK_INO(d_cover_p->d_inode);
	dput(d_mount);
	goto dput_and_out;
    }

    ULOCK_INO(d_cover_p->d_inode);
    dput(d_mount);    /* undo the extra dget() */

    /*
     *  bump the head's refcnt to reflect the mount, and indicate
     *  that the stream is mounted in the flag.  Note that the
     *  flag is cleared in the close process
     */
    lis_atomic_inc(&head->sd_refcnt);
    head->sd_flag |= STRATTACH;

    if (LIS_DEBUG_FATTACH)
	cmn_err( CE_CONT,
		 "fattach( 0x%x, \"%s\" ) done\n", (int)f, path );

dput_and_out:
    dput(d_cover_p);
    ULOCK_INO(i_mount);
    dput(d_cover);         /* undo reference in namei() */
out:
    if (hd_locked && head != NULL)
	lis_spin_unlock(&head->sd_lock) ;	/* unlock strm head */
    return error;
#else
    return -ENOSYS;        /* not implemented for pre-2.1 kernels */
#endif
}

/*
 *  fattach() ioctl wrapper
 *
 *  this is very similar to the syscall wrapper; it differs in the
 *  first parameter (a file * instead of an fd).  It's being used
 *  because lis_check_umem() and lis_copyin_str() don't seem to be
 *  reliable.
 *
 *  Syscall wrapper in syscalls.c
 */
int
lis_ioc_fattach( struct file *f, char *path )
{
    int error;

    error = lis_fattach( f, path );
    return error;
}

/*
 *  fdetach() -
 *
 *  undo an fattach()'s mount.  If the mount reference for the mounted
 *  stream is the last, close the stream.
 *
 *  This routine is written not for the conventional case, where an
 *  fattach mounts a stream on all aliases of a pathname, but to match
 *  our fattach(), which mounts only on the given pathname.  I.e.,
 *  fattach's to links to the pathname, and fattach's to links on
 *  different streams, are not disturbed.
 */
int
lis_fdetach( const char *path )
{
#if defined(KERNEL_2_1)
    struct dentry *d_unmount, *d_uncover, *d_uncover_p;
    struct inode *i_unmount, *i_uncover;
    dev_t dev;
    stdata_t *head;
    cred_t creds;
#ifdef KERNEL_2_3
    struct nameidata nd;
#endif
    int error;

    if (LIS_DEBUG_FATTACH)
	cmn_err( CE_CONT, "fdetach( \"%s\" )\n", path );

#ifdef KERNEL_2_3
    error = user_path_walk( path, &nd );
    if (error)  goto out;
    d_unmount = nd.dentry;
    mntput(nd.mnt);
#else
    d_unmount = namei(path);
    error = PTR_ERR(d_unmount);
    if (IS_ERR(d_unmount))  goto out;
#endif
    d_uncover = d_covers(d_unmount);
    d_uncover_p = dget(d_uncover->d_parent);
    i_uncover = d_uncover->d_inode;
    i_unmount = d_unmount->d_inode;

    if (LIS_DEBUG_FATTACH)
	cmn_err( CE_CONT, "fdetach( \"%s\" ) d_uncover 0x%x d_unmount 0x%x\n",
		 path, (int)d_uncover, (int)d_unmount );

    LOCK_INO(i_unmount);
    
    /*
     *  caller needs write permission to the covered inode
     */
    if ((error = permission( i_uncover, MAY_WRITE )) < 0)
	goto dput_and_out;

    /*
     *  make sure that a stream is mounted here; actually get and
     *  check the stream head, since we need it anyway
     */
    error = -EINVAL;
    if (major(i_unmount->i_dev) != lis_major)
	goto dput_and_out;

    if (d_unmount == d_uncover || d_mounts(d_uncover) != d_unmount)
	goto dput_and_out;

    dev = i_unmount->i_rdev;
    if (!(head = INODE_STR(i_unmount)))
	goto dput_and_out;

    if (LIS_DEBUG_FATTACH)
	cmn_err( CE_CONT,
		 "fdetach() head 0x%x(%d) %s%s inode 0x%x(%d)\n",
		 (int)head, LIS_REFCNT(head),
		 lis_strm_name(head), lis_maj_min_name(head),
		 (int)head->sd_inode, I_COUNT(head->sd_inode) );

    if (head->magic != STDATA_MAGIC || i_unmount != head->sd_inode)
	goto dput_and_out;

    error = -ENOENT;
    LOCK_INO(d_uncover_p->d_inode);
    if (!check_parent(d_uncover_p,d_uncover)) {
	ULOCK_INO(d_uncover_p->d_inode);
	goto dput_and_out;
    }

    /*
     *  our job here is to disassociate the given path from the stream
     *  that covers it.  This means traversing the dentry aliases for
     *  the path's underlying inode, and unlinking each of them.
     *
     *  we undo the stream head ref from fattach() via doclose(), which
     *  will close the dispose of the stream head if apppropriate.
     */
    creds.cr_uid  = current->euid;
    creds.cr_gid  = current->egid;
    creds.cr_ruid = current->uid;
    creds.cr_rgid = current->gid;

    /*
     *  unlink the dentries
     */
    error = dentry_unmount( d_uncover, d_unmount );

    ULOCK_INO(d_uncover_p->d_inode);

    F_CLR(head->sd_flag,STRATTACH);
    lis_doclose( i_unmount, NULL, head, &creds );  /* undo refcnt/flag */
    iput(i_unmount) ;				/* undo fattach's extra igrab */
	    
    if (LIS_DEBUG_FATTACH)
	cmn_err( CE_CONT,
		 "fdetach( \"%s\" ) done\n", path );

dput_and_out:
    dput(d_uncover_p);
    ULOCK_INO(i_unmount);
    dput(d_unmount);    /* undo the namei() above */
out:
    if (LIS_DEBUG_FATTACH)
	if (error) 
	    cmn_err( CE_CONT,
		     "fdetach( \"%s\" ) error %d\n", path, error );
    return error;
#else
    return -ENOSYS;     /* not implemented for pre-2.1 kernels */
#endif
}

/*
 *  lis_fdetach_stream() 
 *
 *  undo any mount(s) (via fattach) on the stream.  The head may be
 *  disposed of in the process, as may a peer stream head if this is
 *  a pipe.
 *
 *  this routine is provided as a means of cleaning up undone fattach
 *  calls.  It should be called when the peer of an fattach'ed pipe end
 *  disappears, but it might also be used before unloading LiS.
 *
 *  unlike most of the related routines, this one returns either 0, to
 *  indicate that it didn't do anything, or > 0, to indicate that it
 *  did.  This is to assist fdetach_all in managing calls to this
 *  routine.
 */
int
lis_fdetach_stream( stdata_t *head )
{
#if defined(KERNEL_2_1)
    struct inode *i_unmount;
    struct list_head *ent;
    int chgd = 0;
    cred_t creds;
    
    if (!head || head->magic != STDATA_MAGIC)
	return 0;

    if (LIS_DEBUG_FATTACH)
	cmn_err( CE_CONT,
		 "fdetach_stream(0x%x) %s%s STRATTACH %s\n",
		 (int)head,
		 lis_strm_name(head), lis_maj_min_name(head),
		 (head->sd_flag & STRATTACH ? "on":"off") );

    /*
     *  we grab the inode to make sure it doesn't disappear from under us
     *  in the loop below; if we can't grab it, we assume it's because it's
     *  not there anymore.
     */
    if (!(i_unmount = igrab(head->sd_inode)))
	return 0;

    LOCK_INO(i_unmount);

    creds.cr_uid  = current->euid;
    creds.cr_gid  = current->egid;
    creds.cr_ruid = current->uid;
    creds.cr_rgid = current->gid;

start_over:
    ent = i_unmount->i_dentry.next;
    while (ent != &(i_unmount->i_dentry)) {
	struct dentry *dum = list_entry( ent, struct dentry, d_alias );
	struct dentry *duc = d_covers(dum);
	struct dentry *ducp = dget(duc->d_parent);

	/*
	 *  See if this dentry is for a (fattach'ed) mount.
	 */
	if (duc != dum &&
	    d_mounts(duc) == dum &&
	    !(permission( duc->d_inode, MAY_WRITE) < 0)) {

	    /*
	     *  unlink the dentries (if the caller has permission)
	     */
	    LOCK_INO(ducp->d_inode);
	    if (check_parent(ducp,duc))
		dentry_unmount( duc, dget(dum) );
	    ULOCK_INO(ducp->d_inode);
	    dput(ducp);
	    dput(dum);			/* from the extra dget just above */
	    if (F_ISSET(head->sd_flag,STRATTACH)) {
		F_CLR(head->sd_flag,STRATTACH);
		lis_doclose( i_unmount, NULL, head, &creds );
	    }

	    chgd++;
	    /*
	     *  the preceding dput(dum) changed the alias list, so we
	     *  will start over instead of taking the link.
	     */
	    goto start_over ;
	} else
	    ent = ent->next;
    }

    ULOCK_INO(i_unmount);

    iput(i_unmount);    /* undo the igrab() above */

    return chgd;
#else
    return 0;           /* not implemented for pre-2.1 kernels */
#endif
}

/*
 *  lis_fdetach_all() -
 *
 *  a wrapper for lis_detach_stream(), to abolish all fattach's (or
 *  at least, those for which the caller has the appropriate privilege).
 */
int
lis_fdetach_all(void)
{
#if defined(KERNEL_2_1)
#if defined(KERNEL_2_3)

    /*
     *  with the list of fattach mounts we maintain, this is easy...
     */
    while (!list_empty(&fattaches)) {
	struct list_head *e = fattaches.next;
	struct vfsmount *mnt = list_entry( e, struct vfsmount, mnt_list );

	lis_fdetach_stream( INODE_STR(mnt->mnt_root->d_inode) );
    }

#else  /* KERNEL_2_1 */
    stdata_t *head = lis_stdata_head, *next;
    long try = 0, lim = lis_stdata_cnt;
    int chgd;

    /*
     *  this may look strange, but we reset the loop if a call to
     *  lis_fdetach_stream() returns positive, since in that case,
     *  one or more heads may have disappeared and thus the list is no
     *  longer to be trusted.  We know we're done when either the
     *  list is empty or we reach the end of the list (having
     *  been returned an error for each head in the list).
     */
    while (head && try++ < lim) {
	next = head->sd_next;
	if ((chgd = lis_fdetach_stream(head)) > 0) {
	    head = lis_stdata_head;
#if 1
	    try = 0;  lim = lis_stdata_cnt;
#endif
	} else
	    head = next;
    }
#endif  /* !KERNEL_2_3 */
#endif  /* KERNEL_2_1 */

    return 0;  /* always */
}

/*
 *  fdetach() ioctl wrapper
 *	      syscall wrapper in syscalls.c
 */
int
lis_ioc_fdetach( char *path )
{
    char *tmp = getname(path);
    int error = PTR_ERR(tmp);

    if (!IS_ERR(tmp)) {
	if (strcmp( tmp, "*" ) == 0)
	    error = lis_fdetach_all();
	else
	    error = lis_fdetach(path);  /* arg must be in user space */
	putname(tmp);
    }

    return error;
}

/*
 *  lis_sendfd() -
 *
 *  we allow either a file pointer or a file descriptor as input here,
 *  in order that "internal" uses may be supported (e.g., connld).  In
 *  fact, the file pointer supercedes the file descriptor, if both are
 *  provided.
 *
 *  in the case that an fp is given, it is used with the assumption
 *  the the reference count has been bumped appropriately.
 */
int lis_sendfd( stdata_t *sendhd, unsigned int fd, struct file *fp )
{
    stdata_t *recvhd = NULL;
    int	      recvhd_locked = 0 ;
    mblk_t *mp;
    strrecvfd_t *sendfd;
    struct file *oldfp;
    int error;

    error = -EPIPE;
    if (!sendhd ||
	!(sendhd->magic == STDATA_MAGIC) ||
	!(recvhd = sendhd->sd_peer) ||
	(sendhd->sd_flag & STRHUP))
	goto not_fifo;

    error = -ENOSR;
    if (!(mp = allocb( sizeof(strrecvfd_t), BPRI_HI )))  goto no_msg;

    /*
     *  if fp is set, we assume it's reference count is set so that
     *  we won't lose it.  Otherwise, we must get the fp corresponding
     *  to fd.
     */
    if (!fp) {
	/*
	 *  get the file pointer corresponding to fd in the current (i.e.,
	 *  the sender's) process.  We do fget() here to hold the returned
	 *  fp for us.
	 */
	error = -EBADF;
	if (!(fp = fget(fd)))  goto bad_file;
	oldfp = fp;
    } else
	oldfp = fp;

    /*
     *  there's one case where we don't want the fp's count bumped;
     *  if the fp is for the receiving file itself, it can't be
     *  closed if the count is high and the FD isn't received.  But
     *  we can't tell for sure if this will happen, so to avoid it,
     *  we make a copy of the file pointer and pass it instead.  We
     *  also pass the original file pointer as r.fp, so it can be
     *  used if it is the receiving file.
     */
    lis_spin_lock(&recvhd->sd_lock) ;		/* lock receiving strm head */
    recvhd_locked = 1 ;
    if (FILE_INODE(fp) == recvhd->sd_inode)  {
	struct file *dfp = lis_get_filp(NULL);
	struct inode *i = FILE_INODE(fp);

	LOCK_INO(i);
	dfp->f_pos   = fp->f_pos;
	dfp->f_flags = fp->f_flags;
	dfp->f_mode  = fp->f_mode;
#if defined(KERNEL_2_1)
	error = -ENOMEM;
	if (!(dfp->f_dentry = lis_d_alloc_root(igrab(i)))) {
	    ULOCK_INO(i);
	    fput(dfp);
	    fput(fp);
	    goto bad_file;
	}
	if (F_ISSET(recvhd->sd_flag,STFIFO))
	    lis_fifo_sendfd_sync( i, dfp );
#else
	FILE_INODE(dfp) = igrab(i);
#endif	    
	SET_FILE_STR(dfp, FILE_STR(fp));

	ULOCK_INO(i);
	
	fput(fp);
	fp = dfp;
    }
    
    /*
     *  OK - set up an M_PASSFP message containing a strrecvfd and
     *  put it in the peer's stream head read queue.
     */
    mp->b_datap->db_type = M_PASSFP;
    sendfd = (strrecvfd_t *) mp->b_rptr;
    sendfd->f.fp  = fp;
    sendfd->uid   = current->euid;
    sendfd->gid   = current->egid;
    sendfd->r.fp  = (FILE_INODE(fp) == recvhd->sd_inode ? oldfp : NULL);
    sendfd->r.hd  = recvhd;
    mp->b_wptr = mp->b_rptr + sizeof(strrecvfd_t);
    (recvhd->sd_rfdcnt)++;

    if (LIS_DEBUG_SNDFD | LIS_DEBUG_IOCTL) {
	cmn_err( CE_CONT,
		 "sendfd( ..., %d, 0x%x ) from %s%s to %s%s",
		 fd, oldfp,
		 lis_strm_name(sendhd), lis_maj_min_name(sendhd),
		 lis_strm_name(recvhd), lis_maj_min_name(recvhd) );
	if (sendfd->r.fp)
	    cmn_err( CE_CONT, " as 0x%x\n", fp );
	else
	    cmn_err( CE_CONT, "\n" );
    }

    /*
     *  the following wakes up a receiver if needed
     */
    if (recvhd_locked)
	lis_spin_unlock(&recvhd->sd_lock) ;
    lis_lockq(recvhd->sd_rq) ;
    lis_strrput( recvhd->sd_rq, mp );
    lis_unlockq(recvhd->sd_rq) ;
    return 0 ;

bad_file:
    freemsg(mp);
no_msg:
not_fifo:
    if (recvhd && recvhd_locked)
	lis_spin_unlock(&recvhd->sd_lock) ;
    return error;
}

/*
 *  the following will be called from flushq() if an M_PASSFP message
 *  is encountered.  This effectively closes the passed file, then
 *  frees the message.
 */
void lis_free_passfp( mblk_t *mp )
{
    strrecvfd_t *sent;
    stdata_t *recvhd;
    int is_a_stream;

    /*
     *  check the message and its contents for validity.
     */
    if (mp->b_datap->db_type != M_PASSFP)  goto not_passfp;
    if (mp->b_wptr - mp->b_rptr < sizeof(strrecvfd_t))  goto not_passfp;

    sent = (strrecvfd_t *) mp->b_rptr;

    is_a_stream = (sent->f.fp->f_op == (&lis_streams_fops));

    recvhd = sent->r.hd;
    lis_spin_lock(&recvhd->sd_lock) ;		/* lock receiving strm head */
    (recvhd->sd_rfdcnt)--;
    lis_spin_unlock(&recvhd->sd_lock) ;

    if (LIS_DEBUG_SNDFD | LIS_DEBUG_CLOSE)
	cmn_err( CE_CONT,
		 "free_passfp(0x%x): %s unreceived %sfile 0x%x\n",
		 mp,
		 (sent->r.fp ? "freeing" : "closing"),
		 (is_a_stream ? "STREAMS " : ""),
		 sent->f.fp );
    
    fput(sent->f.fp);

not_passfp:
    lis_freemsg(mp);
}


int lis_recvfd( stdata_t *recvhd, strrecvfd_t *recv, struct file *fp )
{
    mblk_t *mp;
    strrecvfd_t *sent;
    int error;

    lis_bzero( recv, sizeof(strrecvfd_t) );

    error = -EBADF;
    if (!recvhd || !(recvhd->magic == STDATA_MAGIC))
	goto not_stream;

    /*
     *  we expect the caller to have sync'ed with the sender;
     *  we just fail if no message is waiting
     */
    error = -EAGAIN;
    if (!(mp = getq(recvhd->sd_rq)))
	goto no_msg;

    /*
     *  check the message and its contents for validity.
     */
    if (mp->b_datap->db_type != M_PASSFP)
	goto not_passfp;
    (recvhd->sd_rfdcnt)--;
    if (mp->b_wptr - mp->b_rptr < sizeof(strrecvfd_t))
	goto not_passfp;

    error = -ENFILE;
    if ((recv->f.fd = get_unused_fd()) < 0)
	goto no_fds;

    /*
     *  it's a passed FP - hook up the file that was passed to the new FD
     */
    sent = (strrecvfd_t *) mp->b_rptr;
    recv->uid = sent->uid;
    recv->gid = sent->gid;

    if (sent->r.fp && (sent->r.fp == fp)) {
	if (LIS_DEBUG_SNDFD | LIS_DEBUG_IOCTL)
	    cmn_err( CE_CONT,
		 "recvfd( ..., 0x%x ) S==R, using fp 0x%x, freeing 0x%x\n",
		 sent->f.fp, fp, sent->f.fp );

	fd_install( recv->f.fd, fp );
	(void) fget(recv->f.fd);
	sent->f.fp->f_op = NULL;  /* avoid calling strclose */
	if (F_ISSET(recvhd->sd_flag,STFIFO))
	    lis_fifo_close_sync( FILE_INODE(fp), fp );
	fput(sent->f.fp);
	if (LIS_DEBUG_IOCTL)
	    sent->f.fp = fp;
    } else {
	fd_install( recv->f.fd, sent->f.fp );
    }

    if (LIS_DEBUG_SNDFD | LIS_DEBUG_IOCTL) {
	cmn_err( CE_CONT,
		 "recvfd( ..., 0x%x ) %d at %s%s\n",
		 sent->f.fp, recv->f.fd,
		 lis_strm_name(recvhd), lis_maj_min_name(recvhd) );
    }

    freemsg(mp);  /* we can release the sent message now */

    return 0;     /* OK */

no_fds:
not_passfp:
    putbq( recvhd->sd_rq, mp );   /* put the message back */
no_msg:
not_stream:
    return error;
}


/************************************************************************
*                         Atomic Routines                               *
************************************************************************/

void    lis_atomic_set(lis_atomic_t *atomic_addr, int valu)
{
    atomic_set((atomic_t *)atomic_addr, valu) ;
}

int     lis_atomic_read(lis_atomic_t *atomic_addr)
{
    return(atomic_read(((atomic_t *)atomic_addr))) ;
}

void    lis_atomic_add(lis_atomic_t *atomic_addr, int amt)
{
    atomic_add((amt),((atomic_t *)atomic_addr)) ;
}

void    lis_atomic_sub(lis_atomic_t *atomic_addr, int amt)
{
    atomic_sub((amt),((atomic_t *)atomic_addr)) ;
}

void    lis_atomic_inc(lis_atomic_t *atomic_addr)
{
    atomic_inc(((atomic_t *)atomic_addr)) ;
}

void    lis_atomic_dec(lis_atomic_t *atomic_addr)
{
    atomic_dec(((atomic_t *)atomic_addr)) ;
}

int     lis_atomic_dec_and_test(lis_atomic_t *atomic_addr)
{
    return(atomic_dec_and_test(((atomic_t *)atomic_addr))) ;
}

/************************************************************************
*                         Kernel Semaphores                             *
*************************************************************************
*									*
* These routines are used with doing a down/up on a kernel semaphore.	*
* lis_down/up are used for LiS type semaphores.  Kernel semaphores 	*
* occur in kernel structures, such as inodes.				*
*									*
************************************************************************/
int lis_kernel_down(struct semaphore *sem)
{
    return(down_interruptible(sem)) ;
}

void lis_kernel_up(struct semaphore *sem)
{
    up(sem) ;
}

/************************************************************************
*                         User Space Access                             *
*************************************************************************
*									*
* copy to/from user space.						*
*									*
************************************************************************/

int	lis_copyin(struct file *fp, void *kbuf, const void *ubuf, int len)
{
#if defined(KERNEL_2_1)         /* 2.1 kernel or later */
    return(copy_from_user(kbuf,ubuf,len)?-EFAULT:0) ;
#else                           /* 2.0 kernels */
    return(memcpy_fromfs(kbuf,ubuf,len), 0) ;
#endif

}

int	lis_copyout(struct file *fp, const void *kbuf, void *ubuf, int len)
{
#if defined(KERNEL_2_1)         /* 2.1 kernel or later */
    return(copy_to_user  (ubuf,kbuf,len)?-EFAULT:0) ;
#else                           /* 2.0 kernels */
    return(memcpy_tofs(ubuf,kbuf,len), 0) ;
#endif
}

int lis_check_umem(struct file *fp, int rd_wr_fcn,
		   const void *usr_addr, int lgth)
{
    return(verify_area(rd_wr_fcn,usr_addr,lgth)) ;
}

/************************************************************************
*                         lis_gettimeofday                              *
*************************************************************************
*									*
* A slightly slower version of the kernel routine.			*
*									*
************************************************************************/
void lis_gettimeofday(struct timeval *tv)
{
    return(do_gettimeofday(tv));
}



/*  -------------------------------------------------------------------  */
/*				    Module                               */
#ifdef LINUX

#ifdef LIS_LOADABLE_SUPPORT

/*
 *  lis_loadable_load - load a loadable module.
 *
 *  This function may sleep. On failure it may return a negative errno, or
 *  it may return 0. The module should have been loaded on return from this
 *  function, and the caller should check for this.
 */
int lis_loadable_load(const char *name)
{
	return request_module(name);
}

#endif /* ifdef LIS_LOADABLE_SUPPORT */


int lis_init_module( void )
{
    extern char	*lis_poll_file ;

    printk(
	"==================================================================\n"
	"Linux Streams Subsystem loading...\n");

    lis_print_trace = lis_print_trace_func ;
    lis_major = register_chrdev(0,"streams",&lis_streams_fops);
    if	(lis_major < 0)
    {
	printk("Unable to register Streams Subsystem\n");
	return -EIO;
    }
    /* Initialize every global variable to a default value, if there're
     * modules w/ not-exported globals we should create init_() functions for
     * them and call them from here. */

    lis_spl_init() ;
    lis_spin_lock_init(&lis_runq_lock, "Runq-Lock") ;
    lis_spin_lock_init(&lis_setqsched_lock, "SetQsched-Lock") ;
    lis_spin_lock_init(&lis_task_lock, "Task-Lock") ;
    lis_init_head();

    hook_syscalls();
#ifdef KERNEL_2_3
    {
	int	 err ;

	err = register_filesystem(&lis_file_system_ops) ;
	if (err == 0)
	{
	    lis_mnt = kern_mount(&lis_file_system_ops) ;
	    err = PTR_ERR(lis_mnt) ;
	    if (IS_ERR(lis_mnt))
		unregister_filesystem(&lis_file_system_ops) ;
	    else
		err = 0 ;
	}

	if (err != 0)
	{
	    printk(
	     "Linux Streams Subsystem failed to register its file system (%d).\n",
	     err) ;
	    return(err) ;
	}
    }
#endif

    printk(
	"Linux Streams Subsystem ready.\n"
	"Copyright (c) 1997-2001 David Grothe, et al, http://www.gcom.com\n"
	"Major device number %d.\n"
	"Version %s %s. Compiled for kernel version %s.\n"
	"Using %s %s\n"
	"==================================================================\n",
	lis_major, lis_version, lis_date, lis_kernel_version,
	lis_poll_file, lis_stropts_file );

    return(0);
}

#ifdef MODULE			/* for loadable module support */

/*
 * Magic named routine called by kernel module support code.
 */
int init_module( void )
{
    return(lis_init_module()) ;
}

/*
 * Magic named routine called by kernel module support code.
 */
void cleanup_module( void )
{
   extern void	lis_kill_qsched(void) ;
   /*
    * Make sure no streams modules are running,
    * and de-register devices unregister_netdev (dev);
    */

    restore_syscalls();
    lis_kill_qsched() ;			/* drivers/str/runq.c */
    lis_terminate_head();

#if defined(KERNEL_2_1)			/* 2.2 or 2.4 */
    {
	extern void	lis_pci_cleanup(void) ;
	lis_pci_cleanup() ;
    }
#endif

    unregister_chrdev(lis_major,"streams");
#if defined(KERNEL_2_3)			/* 2.4 */
    unregister_filesystem(&lis_file_system_ops) ;
    if (lis_mnt != NULL)
	kern_umount(lis_mnt) ;
#endif
    printk ("Linux Streams Subsystem removed\n");
}

#endif

/************************************************************************
*                       lis_print_trace_func                            *
*************************************************************************
*									*
* This function is pointed to by the lis_print_trace pointer.  It makes	*
* cmn_err strings come out in the log. 					*
*									*
************************************************************************/
void lis_print_trace_func(char *msg)
{
    printk("%s", msg) ;

} /* lis_print_trace_func */

/************************************************************************
*                          streams_init                                 *
*************************************************************************
*									*
* This routine is called from the Linux main.c.  It registers the 	*
* streams driver and initializes some streams variables.		*
*									*
* No memory is allocated.						*
*									*
************************************************************************/
void	streams_init(void)
{

    lis_init_module() ;		/* register STREAMS with Linux */

} /* streams_init */

/************************************************************************
*                        lis_inc_mod_cnt                                *
*************************************************************************
*									*
* Increment the module counter for STREAMS.  This prevents us from	*
* being unloaded while we have files open.				*
*									*
************************************************************************/
void	lis_inc_mod_cnt(void)
{
    MOD_INC_USE_COUNT ;

} /* lis_inc_mod_cnt */

/************************************************************************
*                          lis_dec_mod_cnt                              *
*************************************************************************
*									*
* Decrement our module use count.					*
*									*
************************************************************************/
void	lis_dec_mod_cnt(void)
{
    MOD_DEC_USE_COUNT ;

} /* lis_dec_mod_cnt */

#endif /* LINUX */

/************************************************************************
*                         Thread Startup                                *
*************************************************************************
*									*
* This routine can be called from other drivers, as well as from LiS	*
* internally, to start up a kernel thread.  It takes care of the	*
* business of shedding user mapped pages and such details.  The idea is	*
* to be able to keep the low-level kernel interactions out of the user's*
* thread code.								*
*									*
* The prototype for this resides in linux-mdep.h and will be automatic-	*
* ally included when you include <sys/stream.h>.			*
*									*
************************************************************************/
/*
 * To lock or not to lock: only 2.2 SMP kernels need to lock
 */
#if defined(__SMP__) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0) \
		     && LINUX_VERSION_CODE <= KERNEL_VERSION(2,3,0)
#include <linux/smp_lock.h>
#define	Klock	lock_kernel()
#define	Kunlock	unlock_kernel()
#else
#define	Klock	
#define	Kunlock	
#endif

/*
 * Some defines for manipulating signals.
 */
#if defined(KERNEL_2_3)
#define	MY_SIGS		current->pending.signal
#define	MY_SIG		MY_SIGS.sig[0]
#define	MY_BLKS		current->blocked
#define	MY_BLKD		MY_BLKS.sig[0]
#elif defined(KERNEL_2_1)
#define MY_SIGS		current->signal
#define MY_SIG		current->signal
#define MY_BLKS		current->blocked
#define MY_BLKD		current->blocked
#endif

/*
 * This is the function that we start up.  It sheds user memory 
 * and calls the user's thread function.  When the user's function
 * returns, it exits.
 *
 * By default these threads run at normal priority.  LiS's own queue runner
 * thread ups its priority to that of a real-time process.
 */
typedef struct
{
    char	 name[sizeof(current->comm)] ;
    int        (*func)(void *func_arg) ;
    void	*func_arg ;

} arg_t ;

int	lis_thread_func(void *argp)
{
    arg_t		*arg = (arg_t *) argp ;
    int		       (*func)(void *) ;
    void		*func_arg ;

    Klock; 				/* get the big kernel lock */
    exit_files(current) ;		/* close all files */
    exit_mm(current) ;			/* detach user pages */
    dput(current->fs->pwd) ;		/* detach from current directory */
    current->fs->pwd = dget(current->fs->root) ;	/* use "/" as pwd */
    current->pgrp = 1 ;			/* detach from parent */
    current->session = 1 ;		/* now owned by "init" */
    current->uid = 0 ;			/* become root */
    current->euid = 0 ;			/* become root */
    current->tty = NULL ;		/* detach from any tty */
    if (current->mm != NULL)
	current->mm->arg_start = current->mm->arg_end = 0;
    strcpy(current->comm, arg->name) ;

    func = arg->func ;
    func_arg = arg->func_arg ;
    FREE(argp) ;			/* don't need args anymore */

    Kunlock ;				/* release big kernel lock */

    return(func(func_arg)) ;		/* enter caller's function */
    					/* without holding big kernel lock */
}

/*
 * Start the thread with "fcn(arg)" as the entry point.  Return the pid for the
 * new process, or < 0 for error.
 */
int	lis_thread_start(int (*fcn)(void *), void *arg, char *name)
{
    arg_t	*argp ;

    argp = ALLOCF(sizeof(*argp), "Thread ") ;
    if (argp == NULL)
	return(-ENOMEM) ;

    if (name != NULL && name[0] != 0)
	strncpy(argp->name, name, sizeof(argp->name)) ;
    else
	strcpy(argp->name, "LiS-Thread") ;

    argp->func = fcn ;
    argp->func_arg = arg ;

    return(kernel_thread(lis_thread_func, (void *) argp, 0)) ;
}

/************************************************************************
*                       Queue Running                                   *
*************************************************************************
*									*
* The following routines are used to manage running the STREAMS queues.	*
* There are two different techniques, selectable via the Configure	*
* script.  One is to use a "bottom half" routine to do the work.  The	*
* other is to use a kernel thread.					*
*									*
* The bottom half technique is faster and more like SVR4 Unix.  It	*
* has in common with SVR4 the fact that the bottom half context prevents*
* STREAMS service procedures from using certain kernel functions which	*
* have to be called from blockable context.				*
*									*
* The kernel thread technique is slower and is more like the multi-	*
* threaded STREAMS of Solaris, except for being the degenerate case	*
* of only one thread.  Its advantage is the STREAMS service procedures	*
* run in blockable context and can use any kernel function.		*
*									*
* The routines lis_setqsched and lis_kill_qsched are called from outside*
* this module.  lis_setqsched is a request to run the queues.  It means	*
* "now would be a good time to _schedule_ the queues to be run".	*
* list_kill_qsched is called from module unload to kill the kernel	*
* thread.  There is a semaphore interlock on this process to make sure	*
* that the process has been killed before unloading the module.		*
*									*
************************************************************************/

#ifdef USE_BH

/*
 * This runs off of the kernel's bottom half mechanism.  This will
 * happen at kernel exit, return from interrupt or from the schedule
 * routine.  We want to make sure that the running of the service
 * queues does not interrupt driver execution via a putnext from
 * the stream head.  So we pay attention to a flag that is set
 * by the stream head whenever it does a putnext down to a streams
 * driver.
 *
 * Note that a putnext upwards from an interrupt routine will not
 * be preempted by this code since it has to unwind to an interrupt
 * exit before the bottom half mechanism runs.
 *
 * I am sure that this will have to be revisited for SMP.
 */
void	lis_tq_runqueues(void *p)	/* bottom half style */
{
    /* printk("lis_tq_runqueues called\n") ; */

    if (!lis_atomic_read(&lis_putnext_flag))
	lis_run_queues(0) ;
    else
	lis_runq_sched = 0 ;		/* ok to reschedule */

    (void) p;
}

void	lis_setqsched(int can_call)		/* bottom half style */
{
    static int			initialized ;
    static struct tq_struct	lis_tq ;

    (void) can_call ;			/* parameter not used */
    if (!initialized)
    {
	initialized	= 1 ;
	lis_tq.next	= NULL ;
	lis_tq.sync	= 0 ;
	lis_tq.routine	= lis_tq_runqueues ;
	lis_tq.data	= 0 ;
	/* printk("lis_setqsched: lis_tq initialized\n") ; */
    }

    if (lis_atomic_read(&lis_queues_running) == 0 && lis_runq_sched == 0)
    {					/* no, then wake up qrun process */
	/* printk("lis_setqsched: queue_task and mark_bh\n") ; */
	lis_runq_sched = 1 ;		/* queues waiting to be run */
	queue_task(&lis_tq, &tq_immediate);
	mark_bh(IMMEDIATE_BH);
    }

} /* lis_setqsched */

/*
 * Called from head code when time to unload LiS module.
 */
void	lis_kill_qsched(void)
{
    /* nothing to kill in bottom half implementation */
}

void	lis_start_qsched(void)
{
    /* nothing to do in bottom half style */
}

#else					/* kernel thread technique */

/*
 * This is the entry point of a kernel thread
 *
 * It appears that when this thread starts up it no longer holds the "big
 * kernel lock" (on 2.2 kernels).  So the first order of business is to acquire
 * that lock, sort of like we got here as the result of a system call.  Running
 * the queues, and calling kernel routines, without holding that lock is
 * troublesome for the 2.2 kernel.
 */
int	lis_thread_runqueues(void *p)
{
    int			 cpu_id = (int) p ;
    int			 sig_cnt  = 0 ;
    unsigned long	 seconds = 0 ;
    lis_semaphore_t	*semp = &lis_runq_sems[cpu_id] ;
    extern char		 lis_print_buffer[] ;
    extern char		*lis_nxt_print_ptr ;

    Klock; 

    printk(KERN_INFO "LiS-RunQ-%s running on CPU %d pid=%d\n",
	    lis_version, cpu_id, current->pid) ;

    current->policy = SCHED_FIFO ;	/* real-time: run when ready */
    current->rt_priority = 50 ;		/* middle value real-time priority */
    sigaddset(&MY_BLKS, SIGTERM) ;	/* inhibit SIGTERM */
#ifdef KERNEL_2_3
    current->cpus_allowed = (1 << cpu_id) ;	/* bind to a CPU */
#endif

    for (;;)
    {
	lis_run_queues(cpu_id) ;	/* run the STREAMS queues */

	/*
	 * When rebooting the kernel sends a TERM signal to all processes
	 * and then a KILL signal.  We want to keep running after the
	 * TERM so that we can close streams from user processes.  We
	 * will go away from a KILL or INT signal so that we can also
	 * be easily killed from the keyboard.
	 */
	if (lis_down(semp) < 0)			/* sleep */
	{					/* interrupted */
	    if (signal_pending(current)		/* killed */
		&& (   sigismember(&MY_SIGS, SIGKILL)
		    || sigismember(&MY_SIGS, SIGINT)
		   )
	       )
		break ;				/* process killed */

	    if (++sig_cnt >= 5)			/* nothing but signals */
	    {
		printk("%s: Signalled: signal=0x%lx blocked=0x%lx, exiting\n",
			current->comm, MY_SIG, MY_BLKD) ;
		sig_cnt = 0 ;
		break ;
	    }
	}
	else
	    sig_cnt = 0 ;

#ifdef KERNEL_2_3
	if (cpu_id != smp_processor_id())
	{
	    static int	msg_cnt ;

	    if (++msg_cnt < 5)
		printk("%s woke up running on CPU %d\n",
			current->comm, smp_processor_id()) ;
	}
#endif
	/*
	 * If there are characters queued up in need of printing, print them if
	 * some time has elapsed.
	 */
	if (cpu_id == 0 && lis_nxt_print_ptr != lis_print_buffer)
	{
	    if (seconds == 0)			/* not timing yet */
		seconds = lis_secs() ;		/* start timing */
	    else
	    if (lis_secs() - seconds > 1)	/* has enough time gone by? */
	    {
		lis_flush_print_buffer() ;
		seconds = 0 ;			/* stop timing */
	    }
	}
    }

    printk(KERN_INFO "%s exiting pid=%d\n", current->comm, current->pid) ;
    lis_sem_destroy(&lis_runq_sems[cpu_id]) ;	/* de-init semaphore */
    lis_runq_pids[cpu_id] = 0 ;			/* process gone */
    lis_up(&lis_runq_kill_sems[cpu_id]) ;	/* OK, we're killed */
    Kunlock ;
    return(0) ;
}

/*
 * Called to start the queue running process.
 */
void	lis_start_qsched(void)
{
    int		cpu ;
    int		ncpus ;
    char	name[16] ;

#ifdef KERNEL_2_3
    ncpus = smp_num_cpus ;
#else
    ncpus = 1 ;
#endif
    lis_num_cpus = smp_num_cpus ;
    lis_spin_lock(&lis_runq_lock) ;		/* one at a time here */
    for (cpu = 0; cpu < ncpus; cpu++)
    {
	if (lis_runq_pids[cpu] > 0)		/* already running */
	    continue ;

	lis_sem_init(&lis_runq_sems[cpu], 0) ;	/* initialize semaphore */
	lis_sem_init(&lis_runq_kill_sems[cpu], 0) ;/* initialize semaphore */

	sprintf(name, "LiS-%s:%u", lis_version, cpu) ;
	lis_runq_pids[cpu] = lis_thread_start(lis_thread_runqueues,
							  (void *)cpu, name) ;
	if (lis_runq_pids[cpu] < 0)		/* failed to fork */
	{
	    printk("lis_start_qsched: %s: lis_thread_start error %d\n",
		    name, lis_runq_pids[cpu]);
	    lis_sem_destroy(&lis_runq_sems[cpu]) ;
	    lis_sem_destroy(&lis_runq_kill_sems[cpu]) ;
	    continue ;
	}

	lis_atomic_inc(&lis_runq_cnt) ;		/* one more running */
    }

    lis_spin_unlock(&lis_runq_lock) ;
}

/*
 * The "can_call" parameter is true if it is OK to just call the queue runner
 * from this routine.  It is false if it is necessary to defer the queue
 * running until later.  Typically qenable() passes 0 and others pass 1.
 */
void	lis_setqsched(int can_call)		/* kernel thread style */
{
    int		runq_cnt ;
    int		queues_running ;
    int		queue_load_thresh ;
    int		req_cnt ;
    int		in_intr = in_interrupt() ;
    int		my_cpu = smp_processor_id() ;
    int		psw ;

    lis_spin_lock_irqsave(&lis_setqsched_lock, &psw) ;
    lis_setqsched_cnts[my_cpu]++ ;	/* keep statistics */
    if (in_intr)
    {
	can_call = 0 ;			/* don't call out from intrs */
	lis_setqsched_isr_cnts[my_cpu]++ ;
    }

    if (lis_atomic_read(&lis_runq_cnt) == 0)
    {					/* no threads running */
	static int	cnt ;
	lis_spin_unlock_irqrestore(&lis_setqsched_lock, &psw) ;
	if (cnt++ < 5)
	    printk("lis_setqsched: No qrun process\n");

	/*
	 * If the runq process is dead we are probably on the way
	 * down for a reboot.  Run the queues from here.
	 */
	if (can_call)
	    lis_run_queues(my_cpu) ;		/* run the STREAMS queues */
	return ;
    }

    /*
     * We keep waking up queue runner threads until they are all running.  We
     * don't want more queue runners actually running than there are queues to
     * be run.
     *
     * runq_cnt is the number of threads that we have started.
     *
     * lis_queues_running is the number of them that are awake and actively
     * processing streams queues.
     *
     * lis_runq_req_cnt is the number of items that are queued for deferred
     * execution by the "queuerun" function.  These are enabled queues, scan
     * queue elements and buffcall callouts pending.  As a heuristic, do not
     * wake up a second queue runner thread until there are more than 4 things
     * queued.  The theory is that it takes some time to wake up and schedule
     * another thread, so there should be enough work queued to make it worth
     * doing.
     *
     * Queueing theory suggests that at 80% utilization the average queue
     * length is expected to be 4.  With 60% utilization you would expect to
     * see 5 or more items queued 5% of the time.  With 45% utilization you
     * would expect to see 5 or more items queued 1% of the time.  So, once
     * there are 5 items queued it is not a bad idea to add another CPU to the
     * mix if one is available.
     *
     * Once we decide to wake up threads, we do so until we have enough threads
     * running to cover the items queued at 4 items per cpu, or until we have
     * awakened the queue runner threads on all available cpus.  This is
     * something of a heuristic so we don't care that some of these numbers
     * may be changing on other cpus while we execute this loop.
     *
     * Because the threads are running at real-time priority we will likely be
     * preempted as soon as we wake up the thread that is bound to the cpu that
     * we are running on.  So we defer that wakeup until last when we have
     * started up all the other cpus.  This also has the effect of waking up
     * the other cpu first for a small number of scheduled queues (say 1).
     *
     * As an efficiency measure, if simply running the queues on our own CPU
     * would suffice to handle the queued entries, then just call out to our
     * own queue runner right from here.  Because of the setting of the
     * "can_call" variable, this means, in effect, when one of the STREAMS
     * system calls is about to exit back to the user.  We don't try to 
     * do this when called from qenable.
     *
     * If we are called from interrupt context we will actually wake up the
     * queue runner on our own CPU (as the last thing we do), but if we are
     * called from background we will just call the run-queues procedure right
     * from here.  This makes single CPU operations more efficient.  Note that
     * if we are called from the queue runner process then the active flag for
     * our CPU will be set and we will neither wake up the process nor call the
     * queue-runner directly, but we might wake up a process on another CPU if
     * it looks like we need more help.
     */
    runq_cnt = lis_atomic_read(&lis_runq_cnt) ;
    queues_running = lis_atomic_read(&lis_queues_running) ;
    queue_load_thresh = 4 * queues_running ;
    req_cnt = lis_atomic_read(&lis_runq_req_cnt) ;
    if (   req_cnt > queue_load_thresh
	&& req_cnt <= queue_load_thresh + 4	/* one more would do it */
	&& !lis_atomic_read(&lis_runq_active_flags[my_cpu])
       )
    {						/* maybe can handle it here */
	lis_spin_unlock_irqrestore(&lis_setqsched_lock, &psw) ;
	if (can_call)
	    lis_run_queues(my_cpu) ;		/* run the queues from here */
	else					/* can't call directly */
	if (lis_atomic_read(&lis_in_syscall) == 0) /* not in a system call */
	    lis_up(&lis_runq_sems[my_cpu]) ;	/* wake up thread on my cpu */
	/* else let syscall exit come back here again */
    }
    else					/* need more help */
    {
	int		cpu ;
	int		qthreshold = 4*queues_running ;
	int		enough = 0 ;

	for (cpu = 0; cpu < runq_cnt; cpu++)/* find a sleeping thread */
	{
	    if (   cpu != my_cpu		/* don't start on my cpu yet */
		&& !lis_atomic_read(&lis_runq_active_flags[cpu]))
	    {					/* sleeping thread */
		lis_up(&lis_runq_sems[cpu]) ;	/* wake up a thread */
		qthreshold += 4 ;			/* increase threshold */
		if (lis_atomic_read(&lis_runq_req_cnt) <= qthreshold)
		{
		    enough = 1 ;		/* enough threads running */
		    break ;
		}
	    }
	}

	if (!enough && !lis_atomic_read(&lis_runq_active_flags[my_cpu]))
	{
	    lis_spin_unlock_irqrestore(&lis_setqsched_lock, &psw) ;
	    if (!can_call)			/* can't just do it */
		lis_up(&lis_runq_sems[my_cpu]) ;/* wake up thread on my cpu */
	    else				/* just do it here */
		lis_run_queues(my_cpu) ;	/* run the queues from here */
	}
	else
	    lis_spin_unlock_irqrestore(&lis_setqsched_lock, &psw) ;
    }

} /* lis_setqsched */

/*
 * Called from head code when time to unload LiS module.
 */
void	lis_kill_qsched(void)
{
    int		cpu ;

    lis_spin_lock(&lis_runq_lock) ;	/* only one CPU does this */
    for (cpu = 0; cpu < smp_num_cpus; cpu++)
    {
	if (lis_runq_pids[cpu] > 0) /* Only stop the kernel thread if running */
	{
	    kill_proc(lis_runq_pids[cpu], SIGKILL, 1) ;
	    lis_down(&lis_runq_kill_sems[cpu]) ;
	    lis_atomic_dec(&lis_runq_cnt) ;		/* one fewer running */
	    lis_sem_destroy(&lis_runq_kill_sems[cpu]) ;	/* de-init semaphore */
	}
    }
    lis_spin_unlock(&lis_runq_lock) ;
}

#endif

/************************************************************************
*                        Process Kill                                   *
************************************************************************/

int      lis_kill_proc(int pid, int sig, int priv)
{
    return(kill_proc(pid, sig, priv)) ;
}
int      lis_kill_pg (int pgrp, int sig, int priv)
{
    return(kill_pg(pgrp, sig, priv)) ;
}

/************************************************************************
*                        Signal Sets                                    *
*************************************************************************
*									*
* We can save and restore signals from the current task structure in	*
* an stdata structure.  We need to do this when closing a stream.  The	*
* close procedure needs to do some semaphore waits, but if the stream	*
* is closing because the process was killed (signalled) then the 	*
* semaphore waits will all fail with EINTR.				*
*									*
************************************************************************/
void lis_clear_and_save_sigs(stdata_t *hd)
{
    sigset_t	*hd_sigs = (sigset_t *) hd->sd_save_sigs ;

#ifdef KERNEL_2_3
    spin_lock_irq(&current->sigmask_lock) ;
    *hd_sigs = current->pending.signal ;
    sigemptyset(&current->pending.signal) ;	/* clear process signals */
    recalc_sigpending(current) ;		/* now none pending */
    spin_unlock_irq(&current->sigmask_lock) ;
#else
    spin_lock_irq(&current->sigmask_lock) ;
    *hd_sigs = current->signal ;
    sigemptyset(&current->signal) ;		/* clear process signals */
    recalc_sigpending(current) ;		/* now none pending */
    spin_unlock_irq(&current->sigmask_lock) ;
#endif
}

void lis_restore_sigs(stdata_t *hd)
{
    sigset_t	*hd_sigs = (sigset_t *) hd->sd_save_sigs ;

#ifdef KERNEL_2_3
    spin_lock_irq(&current->sigmask_lock) ;
    current->pending.signal = *hd_sigs ;
    recalc_sigpending(current) ;		/* now none pending */
    spin_unlock_irq(&current->sigmask_lock) ;
#else
    spin_lock_irq(&current->sigmask_lock) ;
    current->signal = *hd_sigs ;
    recalc_sigpending(current) ;		/* now none pending */
    spin_unlock_irq(&current->sigmask_lock) ;
#endif
}

/************************************************************************
*                      Credentials Manipulation                         *
*************************************************************************
*									*
* These routines copy credentials to/from the task structure.		*
*									*
************************************************************************/
void	lis_task_to_creds(lis_kcreds_t *cp)
{
    int		i ;
    int		psw ;

    lis_spin_lock_irqsave(&lis_task_lock, &psw) ;
    cp->uid = current->uid ;
    cp->euid = current->euid ;
    cp->suid = current->suid ;
    cp->fsuid = current->fsuid ;
    cp->gid = current->gid ;
    cp->egid = current->egid ;
    cp->sgid = current->sgid ;
    cp->fsgid = current->fsgid ;
    cp->cap_effective = current->cap_effective ;
    cp->cap_inheritable = current->cap_inheritable ;
    cp->cap_permitted = current->cap_permitted ;
    cp->ngroups = current->ngroups ;
    for (i = 0; i < current->ngroups; i++)
	cp->groups[i] = current->groups[i] ;
    lis_spin_unlock_irqrestore(&lis_task_lock, &psw) ;
}

void	lis_creds_to_task(lis_kcreds_t *cp)
{
    int		i ;
    int		psw ;

    lis_spin_lock_irqsave(&lis_task_lock, &psw) ;
    current->uid = cp->uid ;
    current->euid = cp->euid ;
    current->suid = cp->suid ;
    current->fsuid = cp->fsuid ;
    current->gid = cp->gid ;
    current->egid = cp->egid ;
    current->sgid = cp->sgid ;
    current->fsgid = cp->fsgid ;
    current->cap_effective = cp->cap_effective ;
    current->cap_inheritable = cp->cap_inheritable ;
    current->cap_permitted = cp->cap_permitted ;
    current->ngroups = cp->ngroups ;
    for (i = 0; i < cp->ngroups; i++)
	current->groups[i] = cp->groups[i] ;
    lis_spin_unlock_irqrestore(&lis_task_lock, &psw) ;
}

/************************************************************************
*                           lis_mknod                                   *
*************************************************************************
*									*
* Like the system call.  "name" is created in the "/dev" directory	*
* with the requested mode and device major/minor.			*
*									*
************************************************************************/
#if 0			/* see rant in dki.h */
int	lis_mknod(char *name, int mode, dev_t dev)
#if defined(KERNEL_2_3)
{
    char		 node_name[100+5] ;	/* "/dev" + name + NUL */
    int			 err = 0 ;
    int			 maj ;
    struct dentry	*dentry ;
    struct nameidata	 nd ;

    if (strlen(name) >= 100)
	return(-ENAMETOOLONG) ;

    if (   S_ISDIR(mode)		/* must not be a directory */
	|| name[0] == '/'		/* must not be in "root" */
       )		
	return(-EPERM) ;

    maj = STR_MAJOR(dev) ;
    if (!LIS_DEVOK(maj))			/* must be streams major */
	return(-ENXIO) ;  			/* was ENOTTY(?) */

    strcpy(node_name, "/dev") ;
    strcat(node_name, name) ;
    if (!path_init(node_name, LOOKUP_PARENT, &nd))
	return(-EPERM) ;			/* not sure of correct errno */

    if ((err = path_walk(node_name, &nd)) != 0)
	return(err) ;

    err = -EEXIST ;
    if (nd.last_type != LAST_NORM)
	goto path_rel_and_return ;

    lis_kernel_down(&nd.dentry->d_inode->i_sem);
    dentry = lookup_hash(&nd.last, nd.dentry);
    err = PTR_ERR(dentry);
    if (IS_ERR(dentry))
	goto unlock_and_return ;

    switch (mode & S_IFMT)
    {
    case 0:
    case S_IFREG:
	/*
	 * We will allow creating a regular file in /dev on the
	 * theory that it will be used for fattach later.  This
	 * may be a bad idea, we'll see.
	 */
	err = vfs_create(nd.dentry->d_inode, dentry, mode);
	break;
    case S_IFCHR:
    case S_IFIFO:
	err = vfs_mknod(nd.dentry->d_inode, dentry, mode, dev) ;
	break ;
    case S_IFDIR:
	err = -EPERM ;
	break ;
    default:
    case S_IFBLK:
    case S_IFSOCK:
	err = -EINVAL ;
	break ;
    }

    dput(dentry);

unlock_and_return:
    lis_kernel_up(&nd.dentry->d_inode->i_sem) ;

path_rel_and_return:
    path_release(&nd) ;

    return(err) ;
}
#else					/* assume 2.2 kernel */
{
    char	 	 node_name[100+5] ;	/* "/dev" + name + NUL */
    int		 	 maj ;
    int			 err ;
    struct dentry	*dentry ;

    if (strlen(name) >= 100)
	return(-ENAMETOOLONG) ;

    if (   S_ISDIR(mode)		/* must not be a directory */
	|| name[0] == '/'		/* must not be in "root" */
       )		
	return(-EPERM) ;

    maj = STR_MAJOR(dev) ;
    if (!LIS_DEVOK(maj))		/* must be a streams major */
	return(-ENXIO) ;

    switch (mode & S_IFMT)
    {
    case 0:
    case S_IFREG:
    case S_IFCHR:
    case S_IFIFO:
	break ;				/* allowable types */
    case S_IFDIR:
	return(-EPERM) ;
	break;
    default:
    case S_IFBLK:
    case S_IFSOCK:
	return(EINVAL) ;
    }

    strcpy(node_name, "/dev") ;
    strcat(node_name, name) ;
    dentry = do_mknod(node_name, mode, dev) ;
    err = PTR_ERR(dentry);
    if (!IS_ERR(dentry))
    {
	dput(dentry) ;
	err = 0 ;
    }

    return(err) ;
}
#endif
#endif				/* if 0 */

/************************************************************************
*                           lis_unlink                                  *
*************************************************************************
*									*
* Remove a name from the "/dev" directory.				*
*									*
************************************************************************/
#if 0			/* see rant in dki.h */
int	lis_unlink(char *name)
#if defined(KERNEL_2_3)
{
    char	 	 node_name[100+5] ;	/* "/dev" + name + NUL */
    int			 err ;

    if (strlen(name) >= 100)
	return(-ENAMETOOLONG) ;

    strcpy(node_name, "/dev") ;
    strcat(node_name, name) ;
}
#else					/* assume 2.2 kernel */
{
    char	 	 node_name[100+5] ;	/* "/dev" + name + NUL */
    int			 err ;
    struct dentry	*dir ;
    struct dentry	*dentry ;

    if (strlen(name) >= 100)
	return(-ENAMETOOLONG) ;

    strcpy(node_name, "/dev") ;
    strcat(node_name, name) ;

    dentry = lookup_dentry(name, NULL, 0) ;
    err = PTR_ERR(dentry) ;
    if (!IS_ERR(dentry))
    {
	dir = lock_parent(dentry) ;
	err = -ENOENT ;
	if (check_parent(dir, dentry))
	    err = vfs_unlink(dir->d_inode, dentry) ;

	unlock_dir(dir) ;
	dput(dentry) ;
    }

    return(err) ;
}
#endif
#endif				/* if 0 */
