/************************************************************************
*                         LiS Locks                                     *
*************************************************************************
*									*
*    Copyright (C) 2000  David Grothe, Gcom, Inc <dave@gcom.com>	*
*									*
* This library is free software; you can redistribute it and/or		*
* modify it under the terms of the GNU Library General Public		*
* License as published by the Free Software Foundation; either		*
* version 2 of the License, or (at your option) any later version.	*
* 									*
* This library is distributed in the hope that it will be useful,	*
* but WITHOUT ANY WARRANTY; without even the implied warranty of	*
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
* Library General Public License for more details.			*
* 									*
* You should have received a copy of the GNU Library General Public	*
* License along with this library; if not, write to the			*
* Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
* Cambridge, MA 02139, USA.						*
*									*
************************************************************************/

#ident "@(#) LiS lislocks.c 1.10 3/12/01"

#include <sys/LiS/linux-mdep.h>
#include <sys/LiS/strmdbg.h>
#include <sys/lislocks.h>
#include <linux/sched.h>
#if defined(KERNEL_2_3)			/* 2.3 kernel or later */
#include <linux/spinlock.h>
#elif defined(KERNEL_2_1)		/* 2.1 and 2.2 kernels */
#include <asm/spinlock.h>
#else					/* earlier kernels */
					/* there will be syntax errors */
#endif


#define FL	char *file, int line

/************************************************************************
*                         SPL Implementation                            *
*************************************************************************
*									*
* We use a global LiS spin lock to implement the SPL functions.		*
*									*
************************************************************************/

lis_spin_lock_t		lis_spl_lock ;		/* for simulating spl fcns */
int			lis_psw0 ;		/* base level psw value */
#define	USE_SPINLOCK	0

#define DCL_l    spinlock_t	*l = (spinlock_t *) lock->spin_lock_mem 

/************************************************************************
*                         SPL Tracking                                  *
*************************************************************************
*									*
* The following structure is used to track SPL usage.			*
*									*
************************************************************************/
typedef struct
{
    int			 type ;
    int			 cpu ;
    void		*addr ;		/* of lock or semaphore */
    void		*tskp ;		/* current task */
    int			 flags ;
    int			 state ;
    char		*file ;
    int			 line ;

} spl_track_t ;

/*
 * Types
 */
#define	TRACK_SPLSTR	1		/* splstr call */
#define	TRACK_SPLX	2		/* splx call */
#define	TRACK_LOCK	3		/* some spinlock call */
#define	TRACK_UNLOCK	4		/* some unlock call */
#define	TRACK_DOWN	5		/* semaphore down */
#define	TRACK_UP	6		/* semaphore up */

#define	SPL_TRACK_SIZE	4096		/* power of 2 */
spl_track_t		 lis_spl_track[SPL_TRACK_SIZE] ;
spl_track_t		*lis_spl_track_ptr = lis_spl_track ;
/*
 * Note: this is a kernel lock, not an LiS lock.
 */
spinlock_t		lis_spl_track_lock = SPIN_LOCK_UNLOCKED ;

#define	LOCK_ENTRY(lock,typ,fil,lin,flgs)				\
    if ( LIS_DEBUG_SPL_TRACE )						\
    {									\
	spl_track_t		*p ;					\
	int			 flags ;				\
									\
	spin_lock_irqsave(&lis_spl_track_lock, flags) ;			\
	p = lis_spl_track_ptr ;						\
	if (++lis_spl_track_ptr >= &lis_spl_track[SPL_TRACK_SIZE])	\
	    lis_spl_track_ptr = lis_spl_track ;				\
	spin_unlock_irqrestore(&lis_spl_track_lock, flags) ;		\
									\
	p->addr = (void *) (lock) ;					\
	p->tskp = (void *) (current) ;					\
	p->file = fil ;							\
	p->line = lin ;							\
	p->type = typ ;							\
	p->flags = flgs ;						\
	p->cpu  = smp_processor_id();					\
	p->state = (lock)->nest ;					\
    }

#define LOCK_EXIT(lock,typ,fil,lin,flgs)				\
    if ( LIS_DEBUG_SPL_TRACE )						\
    {									\
	spl_track_t		*p ;					\
	int			 flags ;				\
									\
	spin_lock_irqsave(&lis_spl_track_lock, flags) ;			\
	p = lis_spl_track_ptr ;						\
	if (++lis_spl_track_ptr >= &lis_spl_track[SPL_TRACK_SIZE])	\
	    lis_spl_track_ptr = lis_spl_track ;				\
	spin_unlock_irqrestore(&lis_spl_track_lock, flags) ;		\
									\
	p->addr = (void *) (lock) ;					\
	p->tskp = (void *) (current) ;					\
	p->file = fil ;							\
	p->line = lin ;							\
	p->type = typ ;							\
	p->flags = flgs ;						\
	p->cpu  = smp_processor_id();					\
	p->state = (lock)->nest ;					\
    }

#define	SEM_ENTRY(sem,typ,fil,lin,flgs)					\
    if ( LIS_DEBUG_SPL_TRACE )						\
    {									\
	spl_track_t		*p ;					\
	int			 flags ;				\
									\
	spin_lock_irqsave(&lis_spl_track_lock, flags) ;			\
	p = lis_spl_track_ptr ;						\
	if (++lis_spl_track_ptr >= &lis_spl_track[SPL_TRACK_SIZE])	\
	    lis_spl_track_ptr = lis_spl_track ;				\
	spin_unlock_irqrestore(&lis_spl_track_lock, flags) ;		\
									\
	p->addr = (void *) (sem) ;					\
	p->tskp = (void *) (current) ;					\
	p->file = fil ;							\
	p->line = lin ;							\
	p->type = typ ;							\
	p->flags = flgs ;						\
	p->cpu  = smp_processor_id();					\
	p->state = atomic_read(&((struct semaphore *) &sem->sem_mem)->count); \
    }

#define SEM_EXIT(sem,typ,fil,lin,flgs)					\
    if ( LIS_DEBUG_SPL_TRACE )						\
    {									\
	spl_track_t		*p ;					\
	int			 flags ;				\
									\
	spin_lock_irqsave(&lis_spl_track_lock, flags) ;			\
	p = lis_spl_track_ptr ;						\
	if (++lis_spl_track_ptr >= &lis_spl_track[SPL_TRACK_SIZE])	\
	    lis_spl_track_ptr = lis_spl_track ;				\
	spin_unlock_irqrestore(&lis_spl_track_lock, flags) ;		\
									\
	p->addr = (void *) (sem) ;					\
	p->tskp = (void *) (current) ;					\
	p->file = fil ;							\
	p->line = lin ;							\
	p->type = typ ;							\
	p->flags = flgs ;						\
	p->cpu  = smp_processor_id();					\
	p->state = atomic_read(&((struct semaphore *) &sem->sem_mem)->count); \
    }

/************************************************************************
*                         Prototypes                                    *
************************************************************************/

extern unsigned long lis_hitime(void);  /* usec res; 64s cycle */


/************************************************************************
*                         lis_print_spl_track                           *
*************************************************************************
*									*
* Print out the spl tracking table.					*
*									*
************************************************************************/
void	lis_print_spl_track(void)
{
    spl_track_t		*p ;
    char		*typep ;
    int			 n ;

    printk("lis_print_spl_track: %lx lis_spl_track_ptr=%lx\n",
		(long)lis_spl_track, (long)lis_spl_track_ptr) ;

    for (p = lis_spl_track_ptr, n = SPL_TRACK_SIZE; n--;)
    {
	if (p->type != 0)
	{
	    switch (p->type)
	    {
	    case TRACK_SPLSTR:	typep = "splstr" ; break ;
	    case TRACK_SPLX:	typep = "splx  " ; break ;
	    case TRACK_LOCK:	typep = "lock  " ; break ;
	    case TRACK_UNLOCK:	typep = "unlock" ; break ;
	    case TRACK_DOWN:	typep = "down  " ; break ;
	    case TRACK_UP:	typep = "up    " ; break ;
	    default:		typep = "Unkwn " ; break ;
	    }

	    printk("%s State=%d Flgs=%03x %s #%u\n",
		    typep,
		    p->state,
		    p->flags & 0x200,
		    p->file, p->line) ;
	}

	if (++p >= &lis_spl_track[SPL_TRACK_SIZE])
	    p = lis_spl_track ;
    }

    printk("lis_print_spl_track: end of output\n") ;
    
} /* lis_print_spl_track */

/************************************************************************
*                             lis_splstr                                *
*************************************************************************
*									*
* Disable interrupts, return the previous state.			*
*									*
************************************************************************/
int	lis_splstr_fcn(char *file, int line)
{
    int			 prev ;

    lis_spin_lock_irqsave_fcn(&lis_spl_lock, &prev, file, line) ;
    return(prev) ;

} /* lis_splstr */

/************************************************************************
*                             lis_splx                                  *
*************************************************************************
*									*
* Restore state.							*
*									*
************************************************************************/
void	lis_splx_fcn(int x, char *file, int line)
{
    lis_spin_unlock_irqrestore_fcn(&lis_spl_lock, &x, file, line) ;

} /* lis_splx */

/************************************************************************
*                             lis_spl0                                  *
*************************************************************************
*									*
* Go all the way back to base level.					*
*									*
************************************************************************/
void	lis_spl0_fcn(char *file, int line)
{
    while (lis_own_spl())
	lis_spin_unlock_irqrestore_fcn(&lis_spl_lock, &lis_psw0, file, line) ;

} /* lis_spl0 */

/************************************************************************
*                            lis_spl_init                               *
*************************************************************************
*									*
* Called only from LiS init routine.					*
*									*
************************************************************************/
void	lis_spl_init(void)
{
    lis_spin_lock_init(&lis_spl_lock, "LiS_SPL_Lock") ;
    lis_psw0 = lis_splstr() ;		/* capture base-level psw */
    lis_splx(lis_psw0) ;
}

/************************************************************************
*                            lis_own_spl                                *
*************************************************************************
*									*
* Return true if this task owns the global lis_spin_lock.		*
*									*
************************************************************************/
int	lis_own_spl(void)
{
    return (lis_spin_is_locked(&lis_spl_lock) && lis_spl_lock.taskp == current);
}

/************************************************************************
*                         Spin Locks                                    *
*************************************************************************
*									*
* These are the routines that interface to the kernel's spin lock 	*
* mechanism.								*
*									*
* The 2.2 kernel has "pretend" spin locks.  If there is no spin lock	*
* debugging enabled they don't even maintain the state of the lock!	*
* That means that we can't test it to return a sensible value.  So for	*
* 2.2 we just return 0 (not locked).  If you need to turn on kernel	*
* spin lock debugging then just uncomment out the line of code here	*
* and we will return a value based upon the state of the lock.		*
*									*
************************************************************************/

int	lis_spin_is_locked_fcn(lis_spin_lock_t *lock, FL)
{
    DCL_l ;

    (void) file; (void) line ;
    (void) l ;				/* compiler happiness in 2.2 */
#if defined(KERNEL_2_3)			/* 2.3 kernel or later */
    return(spin_is_locked(l)) ;
#elif defined(KERNEL_2_1)		/* 2.1 and 2.2 kernels */
    /* return((*((volatile int *) (&l->lock))) == 1) ; */
    return(0) ;				/* see comment above */
#else					/* earlier kernels */
    return(0) ;
#endif
}

void    lis_spin_lock_fcn(lis_spin_lock_t *lock, FL)
{
    int		 prev ;
    DCL_l ;

    save_flags(prev) ;
    lock->spinner_file = file ;
    lock->spinner_line = line ;
    if ((void *) current != lock->taskp)
    {
	spin_lock(l) ;
	lock->taskp = (void *) current ;
	lock->owner_file = file ;
	lock->owner_line = line ;
    }
    lis_atomic_inc(&lock->nest) ;
    LOCK_ENTRY(lock,TRACK_LOCK,file,line,prev)
}

void    lis_spin_unlock_fcn(lis_spin_lock_t *lock, FL)
{
    int		 prev ;
    DCL_l ;

    (void) l ;				/* avoid warning in non-SMP case */
    save_flags(prev) ;

    LOCK_EXIT(lock,TRACK_UNLOCK,file,line,prev)
    if (lis_atomic_read(&lock->nest) > 0)
    {
	lis_atomic_dec(&lock->nest) ;
	if (lis_atomic_read(&lock->nest) == 0)
	{
	    lock->taskp = NULL ;
	    lock->unlocker_file = file ;
	    lock->unlocker_line = line ;
	    spin_unlock(l) ;
	}
    }
}

int     lis_spin_trylock_fcn(lis_spin_lock_t *lock, FL)
{
    int		ret ;
    DCL_l ;

    (void) l ;				/* avoid warning in non-SMP case */
    lock->spinner_file = file ;
    lock->spinner_line = line ;
    if ((void *) current != lock->taskp)
    {
	if ((ret = spin_trylock(l)) != 0)
	{
	    lis_atomic_inc(&lock->nest) ;
	    lock->taskp = (void *) current ;
	    lock->owner_file = file ;
	    lock->owner_line = line ;
	}
	return(ret) ;
    }
    return(1) ;				/* already held */
}

void    lis_spin_lock_irq_fcn(lis_spin_lock_t *lock, FL)
{
    int		 prev ;
    DCL_l ;

    (void) l ;				/* compiler happiness in 2.2 */
    save_flags(prev) ;

    lock->spinner_file = file ;
    lock->spinner_line = line ;
    if ((void *) current != lock->taskp)
    {
	spin_lock_irq(l) ;
	lock->taskp = (void *) current ;
	lock->owner_file = file ;
	lock->owner_line = line ;
    }
    lis_atomic_inc(&lock->nest) ;
    LOCK_ENTRY(lock,TRACK_LOCK,file,line,prev)
}

void    lis_spin_unlock_irq_fcn(lis_spin_lock_t *lock, FL)
{
    int		 prev ;
    DCL_l ;

    (void) l ;				/* avoid warning in non-SMP case */
    save_flags(prev) ;

    LOCK_EXIT(lock,TRACK_UNLOCK,file,line,prev)
    if (lis_atomic_read(&lock->nest) > 0)
    {
	lis_atomic_dec(&lock->nest) ;
	if (lis_atomic_read(&lock->nest) == 0)
	{
	    lock->taskp = NULL ;
	    lock->unlocker_file = file ;
	    lock->unlocker_line = line ;
	    spin_unlock_irq(l) ;
	}
    }
}

void    lis_spin_lock_irqsave_fcn(lis_spin_lock_t *lock, int *flags, FL)
{
    int		 prev ;
    DCL_l ;

    (void) l ;				/* compiler happiness in 2.2 */
    save_flags(prev) ;

    lock->spinner_file = file ;
    lock->spinner_line = line ;
    if ((void *) current != lock->taskp)
    {
	spin_lock_irqsave(l, (*flags)) ;
	lock->taskp = (void *) current ;
	lock->owner_file = file ;
	lock->owner_line = line ;
    }
    lis_atomic_inc(&lock->nest) ;
    LOCK_ENTRY(lock,TRACK_LOCK,file,line,prev)
}

void    lis_spin_unlock_irqrestore_fcn(lis_spin_lock_t *lock, int *flags, FL)
{
    int		 prev ;
    DCL_l ;

    (void) l ;				/* avoid warning in non-SMP case */
    save_flags(prev) ;

    LOCK_EXIT(lock,TRACK_UNLOCK,file,line,prev)
    if (lis_atomic_read(&lock->nest) > 0)
    {
	lis_atomic_dec(&lock->nest) ;
	if (lis_atomic_read(&lock->nest) == 0)
	{
	    lock->taskp = NULL ;
	    lock->unlocker_file = file ;
	    lock->unlocker_line = line ;
	    spin_unlock_irqrestore(l, (*flags)) ;
	}
    }
}

void    lis_spin_lock_init_fcn(lis_spin_lock_t *lock, const char *name, FL)
{
    DCL_l ;

    (void) l ;				/* avoid warning in non-SMP case */
    memset((void *)lock, 0, sizeof(lock)) ;
    if (name == NULL) name = "Spin-Lock" ;

    lock->name 		= (char *) name ;
    lock->spinner_file  = "Initialized" ;
    lock->owner_file    = lock->spinner_file ;
    lock->unlocker_file = lock->spinner_file ;
    spin_lock_init(l) ;			/* kernel's init function */
}

/************************************************************************
*                        Semaphores                                     *
*************************************************************************
*									*
* Subroutines here to insulate from kernel version differences.		*
*									*
************************************************************************/

/*
 * Initialize a semaphore
 */
#if defined(KERNEL_2_1)			/* 2.1 kernel or later */
# define SET_SEM_COUNT(semp,cnt)	sema_init((semp), (cnt))
#else					/* 2.0 kernel */
# ifdef ATOMIC_INIT
# define SET_SEM_COUNT(semp,cnt)	atomic_set(&((semp)->count), (cnt))
# else
# define SET_SEM_COUNT(semp,cnt)	(semp)->count = (cnt)
# endif
#endif

void	lis_up_fcn(lis_semaphore_t *lsem, FL)
{
    struct semaphore	*sem = (struct semaphore *) lsem->sem_mem ;

    SEM_EXIT(lsem,TRACK_UP,file,line,0) ;
    lsem->upper_file = file ;		/* most recent "up" */
    lsem->upper_line = line ;
    lsem->taskp      = NULL ;
    up(sem) ;
}

int	lis_down_fcn(lis_semaphore_t *lsem, FL)
{
    struct semaphore	*sem = (struct semaphore *) lsem->sem_mem ;
    int			 ret ;

    lsem->downer_file = file ;		/* most recent "down" */
    lsem->downer_line = line ;
    ret = down_interruptible(sem) ;
    if (ret == 0)
    {
	SEM_ENTRY(lsem,TRACK_DOWN,file,line,0) ;
	lsem->taskp = (void *) current ;
	lsem->owner_file = file ;	/* current owner */
	lsem->owner_line = line ;
    }
    else
    {
	SEM_ENTRY(lsem,TRACK_DOWN,file,line,ret) ;
	lsem->owner_file = "Error" ;
	lsem->owner_line = ret ;
    }

    return(ret) ;
}

void	lis_sem_init(lis_semaphore_t *lsem, int count)
{
    struct semaphore	*sem = (struct semaphore *) lsem->sem_mem ;
    static lis_semaphore_t lis_sem_template ;	/* blank semaphore */
    *lsem = lis_sem_template ;			/* blank semaphore */
    SET_SEM_COUNT(sem,count);
    lsem->owner_file = "Initialized" ;
}


void	lis_sem_destroy(lis_semaphore_t *lsem)
{
    static lis_semaphore_t lis_sem_template ;	/* blank semaphore */
    *lsem = lis_sem_template ;			/* blank semaphore */
    lsem->owner_file = "De-Initialized" ;
}
