/*                               -*- Mode: C -*- 
 * syscalls.c --- 
 * Author          : Denis Froschauer
 * Created On      : 
 * Last Modified By: David Grothe (dave@gcom.com)
 * RCS Id          : $Id: syscalls.c,v 1.4 1996/01/23 19:40:13 dave Exp $
 * ----------------______________________________________________
 *   Copyright (C) 1995  Denis Froschauer
 *   Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Library General Public
 *    License as published by the Free Software Foundation; either
 *    version 2 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this library; if not, write to the Free
 *    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    You can reach me by email to
 *    100741.1151@compuserve.com
 */


#ident "@(#) LiS syscalls.c 2.9 3/12/01 21:14:31 "

#ifdef LINUX

#include <sys/stream.h>
#include <sys/stropts.h>	/* streams config symbols are here */
#include <sys/poll.h>		/* usr structs */
#include <linux/unistd.h>	/* kernel system call definitions */
#include <sys/osif.h>

/*
 * Do not evaluate the following if we are only doing
 * dependencies.
 */
#if !defined(DEP) && !defined(__NR_putpmsg)

# ifndef __NR_streams2
# error "unistd.h does not define __NR_putpmsg or __NR_streams2"
# else		/* Linus defined __NR_streams2 rather than __NR_putpmsg */
# define __NR_putpmsg	__NR_streams2
# endif

#endif

#define	NR_putpmsg_2_1	189	/* 2.1 and onward */


#if !defined(DEP) && !defined(__NR_getpmsg)

# ifndef __NR_streams1
# error "unistd.h does not define __NR_getpmsg or __NR_streams1"
# else		/* Linus defined __NR_streams1 rather than __NR_getpmsg */
# define __NR_getpmsg	__NR_streams1
# endif

#endif

#define	NR_getpmsg_2_1	188	/* 2.1 and onward */

#if 0
/*
 * In head.c
 */
extern int
lis_strputpmsg(struct inode *i, struct file *fp, 
	       void *ctlp, void *datp, int band, int flags) ;
extern int
lis_strgetpmsg(struct inode *i, struct file *fp, 
	       void *ctlp, void *datp, int *bandp, int *flagsp,
	       int doit) ;
#endif

/*
 * In poll.c
 */
#if defined(PORTABLE_POLL)
extern int	lis_poll(struct poll *fds, int nfds, int time_out) ;
#endif

/*
 * To lock or not to lock: only SMP kernels need to lock
 */
#if defined(KERNEL_2_1) && defined(__SMP__)
#include <linux/smp_lock.h>
#define	Klock	lock_kernel()
#define	Kunlock	unlock_kernel()
#else
#define	Klock	
#define	Kunlock	
#endif

/*
 *  the following are for saving syscall entry points
 */
#ifdef __NR_spipe
static void *default_sys_pipe        = NULL;
#endif
static void *default_sys_putpmsg     = NULL;
static void *default_sys_getpmsg     = NULL;
#ifdef i386
static void *default_sys_putpmsg_2_1 = NULL;
static void *default_sys_getpmsg_2_1 = NULL;
#endif
#ifdef __NR_spipe
static void *default_sys_spipe       = NULL;
#endif
#ifdef __NR_fattach
static void *default_sys_fattach     = NULL;
#endif
#ifdef __NR_fdetach
static void *default_sys_fdetach     = NULL;
#endif

extern void     *sys_call_table[];

#ifdef __NR_spipe
/*
 *  these simplify notation when intercepting syscall entry points
 */
typedef asmlinkage int (*syscall_pipe_t)(unsigned long *);
#endif

/*  -------------------------------------------------------------------  */
/* sys_poll -  poll() syscall entry point
 *
 * STATUS: complete, tested
 */
#if defined(PORTABLE_POLL)
asmlinkage int
sys_poll(void *ubuff, ulong n, int tmout)
{
    return(lis_poll((struct poll *) ubuff, n, tmout)) ;

}/*sys_poll*/
#endif

/*  -------------------------------------------------------------------  */
/* sys_putpmsg -  putpmsg syscall entry point
 *
 * STATUS: complete, tested
 */
asmlinkage int
sys_putpmsg(int fd,
	    struct strbuf *ctl,
	    struct strbuf *dat,
	    int band,
	    int flags)
{
    struct file		* file;
    struct inode	* inode;
    int			  ret ;

    /*
     * The following checks are patterned after sys_write
     */
    Klock ;
#if defined(KERNEL_2_1)
    if (   (file = fget(fd)) == NULL
	|| file->f_dentry == NULL
	|| (inode = file->f_dentry->d_inode) == NULL
       )
    {
	if (file != NULL)
	    fput(file) ;

	Kunlock ;
	return -EBADF;
    }
#else
    if (fd>=NR_OPEN || !(file=current->files->fd[fd]) || !(inode=file->f_inode))
    {
	Kunlock ;
	return -EBADF;
    }
#endif
    if (!(file->f_mode & 2))
    {
#if defined(KERNEL_2_1)
	fput(file) ;
#endif
	Kunlock ;
	return -EBADF;
    }
    if (!file->f_op || !file->f_op->write)
    {
#if defined(KERNEL_2_1)
	fput(file) ;
#endif
	Kunlock ;
	return -EINVAL;
    }

    LOCK_INO(inode);
    ret = lis_strputpmsg(inode,file,ctl,dat,band,flags);
    ULOCK_INO(inode);

#if defined(KERNEL_2_1)
    fput(file) ;
#endif
    Kunlock ;
    return(ret) ;

}/*sys_putpmsg*/

/*  -------------------------------------------------------------------  */
/* sys_getpmsg -  getpmsg syscall entry point
 *
 * STATUS: complete, tested
 */
asmlinkage int
sys_getpmsg(int fd,
	    struct strbuf *ctl,
	    struct strbuf *dat,
	    int *bandp,
	    int *flagsp)
{
    struct file		* file;
    struct inode	* inode;
    int			  ret ;

    /*
     * The following checks are patterned after sys_read
     */
    Klock ;
#if defined(KERNEL_2_1)
    if (   (file = fget(fd)) == NULL
	|| file->f_dentry == NULL
	|| (inode = file->f_dentry->d_inode) == NULL
       )
    {
	if (file != NULL)
	    fput(file) ;

	Kunlock ;
	return -EBADF;
    }
#else
    if (fd>=NR_OPEN || !(file=current->files->fd[fd]) || !(inode=file->f_inode))
    {
	Kunlock ;
	return -EBADF;
    }
#endif
    if (!(file->f_mode & 1))
    {
#if defined(KERNEL_2_1)
	fput(file) ;
#endif
	Kunlock ;
	return -EBADF;
    }
    if (!file->f_op || !file->f_op->read)
    {
#if defined(KERNEL_2_1)
	fput(file) ;
#endif
	Kunlock ;
	return -EINVAL;
    }

    LOCK_INO(inode);
    ret = lis_strgetpmsg(inode,file,ctl,dat,bandp,flagsp,1);
    ULOCK_INO(inode);

#if defined(KERNEL_2_1)
    fput(file) ;
#endif
    Kunlock ;
    return(ret) ;

}/*sys_getpmsg*/

/*
 *  pipe() system call wrapper
 */
asmlinkage int
lis_sys_pipe( unsigned int *fildes )
{
    unsigned int fd[2];
    int error;

    Klock;
    error = lis_pipe(fd);
    Kunlock;

    if (!error) {
	if (copy_to_user(fildes, fd, 2*sizeof(unsigned int)))
	    error = -EFAULT;
    }
    return error;
}

/*
 *  fattach() system call wrapper
 */
asmlinkage int
lis_sys_fattach( int fd, const char *path )
{
    struct file *f;
    int error;

    Klock;
    error = -EBADF;
    if ((f = fcheck(fd)))
	error = lis_fattach( f, path );
    Kunlock;

    return error;
}

/*
 *  fdetach() system call wrapper
 */
asmlinkage int
lis_sys_fdetach( const char *path )
{
    char *tmp = getname(path);
    int error = PTR_ERR(tmp);
    extern int lis_fdetach_all(void) ;

    if (!IS_ERR(tmp)) {
	Klock;
	if (strcmp( tmp, "*" ) == 0)
	    error = lis_fdetach_all();
	else
	    error = lis_fdetach(path);  /* arg must be in user space */
	Kunlock;
	putname(tmp);
    }

    return error;
}

void
hook_syscalls(void)
{
#ifdef __NR_spipe
    /*
     *  get some real entry points, for testing/passing through
     */
    default_sys_pipe  = (void *) sys_call_table[__NR_pipe];
#endif
    /*
     *  save the old entry points
     */
    default_sys_putpmsg = (void *) sys_call_table[__NR_putpmsg];
    default_sys_getpmsg = (void *) sys_call_table[__NR_getpmsg];
#ifdef i386
    default_sys_putpmsg_2_1 = (void *) sys_call_table[NR_putpmsg_2_1];
    default_sys_getpmsg_2_1 = (void *) sys_call_table[NR_getpmsg_2_1];
#endif
#ifdef PORTABLE_POLL
    default_sys_poll    = (void *) sys_call_table[__NR_poll];
#endif
#ifdef __NR_spipe
    default_sys_spipe   = (void *) sys_call_table[__NR_spipe];
#endif
#ifdef __NR_fattach
    default_sys_fattach = (void *) sys_call_table[__NR_fattach];
#endif
#ifdef __NR_fdetach
    default_sys_fdetach = (void *) sys_call_table[__NR_fdetach];
#endif

        /* 
     *  now hook our entry points
        */
    sys_call_table[__NR_putpmsg]	= (void *)sys_putpmsg;
    sys_call_table[__NR_getpmsg]	= (void *)sys_getpmsg;
#ifdef i386
        sys_call_table[  NR_putpmsg_2_1]	= (void *)sys_putpmsg;
        sys_call_table[  NR_getpmsg_2_1]	= (void *)sys_getpmsg;
#endif
#if defined(PORTABLE_POLL)
        sys_call_table[__NR_poll]		= (void *)sys_poll;
#endif
#ifdef __NR_spipe
    sys_call_table[__NR_spipe]          = (void *)lis_sys_pipe;
#endif
#ifdef __NR_fattach
    sys_call_table[__NR_fattach]        = (void *)lis_sys_fattach;
#endif
#ifdef __NR_fdetach
    sys_call_table[__NR_fdetach]        = (void *)lis_sys_fdetach;
#endif

        return;
}

void
restore_syscalls(void)
{
    sys_call_table[__NR_putpmsg]     = (void *)default_sys_putpmsg;
    sys_call_table[__NR_getpmsg]     = (void *)default_sys_getpmsg;
#ifdef i386
    sys_call_table[  NR_putpmsg_2_1] = (void *)default_sys_putpmsg_2_1;
    sys_call_table[  NR_getpmsg_2_1] = (void *)default_sys_getpmsg_2_1;
#endif
#if defined(PORTABLE_POLL)
    sys_call_table[__NR_poll]	     = (void *)default_sys_poll;
#endif
#ifdef __NR_spipe
    sys_call_table[__NR_spipe]       = (void *) default_sys_spipe;
#endif
#ifdef __NR_fattach
    sys_call_table[__NR_fattach]     = (void *) default_sys_fattach;
#endif
#ifdef __NR_fdetach
    sys_call_table[__NR_fdetach]     = (void *) default_sys_fdetach;
#endif

        return;
}

#endif				/* LINUX */
